#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

from idaes.core.surrogate.pysmo.sampling import (
    LatinHypercubeSampling,
    UniformSampling,
    HammersleySampling,
    CVTSampling,
    SamplingMethods,
    FeatureScaling,
)
import idaes.logger as idaeslog
import numpy as np
import pandas as pd
import pyomo.common.unittest as unittest
import pytest


class FeatureScalingTestCases(unittest.TestCase):
    """
    test_data_scaling_minmax_01: Test behaviour when input is a numpy array and with 1D array.
    test_data_scaling_minmax_02: Test behaviour when input is a numpy array and with 2D array.
    test_data_scaling_minmax_03: Test behaviour when input is a numpy array and with 3D array.
    test_data_scaling_minmax_04: Test behaviour when input is a numpy array and with 3D array with a variable is constant.
    test_data_scaling_minmax_05: Test behaviour list input TypeError

    test_data_scaling_minmax_06: Test behaviour when input is a Pandas DataFrame and with 1D array.
    test_data_scaling_minmax_07: Test behaviour when input is a Pandas DataFrame and with 2D array.
    test_data_scaling_minmax_08: Test behaviour when input is a Pandas DataFrame and with 3D array.
    test_data_scaling_minmax_09: Test behaviour when input is a Pandas DataFrame and with 3D array with a variable is constant.

    test_data_unscaling_minmax_01: Test behaviour when input is a numpy array and with 1D array.
    test_data_unscaling_minmax_02: Test behaviour when input is a numpy array and with 2D array.
    test_data_unscaling_minmax_03: Test behaviour when input is a numpy array and with 3D array.
    test_data_unscaling_minmax_04: Test behaviour when input is a numpy array and with 3D array with a variable is constant.

    test_data_unscaling_minmax_05: Test behaviour IndexError when input array size > array size
    test_data_unscaling_minmax_06: Test behaviour IndexError when input array size < array size

    """

    def setUp(self):
        input_array_np_1d = np.array([[0], [1], [2], [3], [4], [5], [6], [7], [8], [9]])
        input_array_np_2d = np.array(
            [
                [0, 1],
                [1, 4],
                [2, 9],
                [3, 16],
                [4, 25],
                [5, 36],
                [6, 49],
                [7, 64],
                [8, 81],
                [9, 100],
            ]
        )
        input_array_np_3d = np.array(
            [
                [0, 10, 11],
                [1, 11, 15],
                [2, 12, 21],
                [3, 13, 29],
                [4, 14, 39],
                [5, 15, 51],
                [6, 16, 65],
                [7, 17, 81],
                [8, 18, 99],
                [9, 19, 119],
            ]
        )
        input_array_np_3d_constant = np.array(
            [
                [0, 10, 11],
                [1, 10, 14],
                [2, 10, 19],
                [3, 10, 26],
                [4, 10, 35],
                [5, 10, 46],
                [6, 10, 59],
                [7, 10, 74],
                [8, 10, 91],
                [9, 10, 110],
            ]
        )

        input_array_pd_1d = pd.DataFrame(
            [[0], [1], [2], [3], [4], [5], [6], [7], [8], [9]]
        )
        input_array_pd_2d = pd.DataFrame(
            [
                [0, 1],
                [1, 4],
                [2, 9],
                [3, 16],
                [4, 25],
                [5, 36],
                [6, 49],
                [7, 64],
                [8, 81],
                [9, 100],
            ]
        )
        input_array_pd_3d = pd.DataFrame(
            [
                [0, 10, 11],
                [1, 11, 15],
                [2, 12, 21],
                [3, 13, 29],
                [4, 14, 39],
                [5, 15, 51],
                [6, 16, 65],
                [7, 17, 81],
                [8, 18, 99],
                [9, 19, 119],
            ]
        )
        input_array_pd_3d_constant = pd.DataFrame(
            [
                [0, 10, 11],
                [1, 10, 14],
                [2, 10, 19],
                [3, 10, 26],
                [4, 10, 35],
                [5, 10, 46],
                [6, 10, 59],
                [7, 10, 74],
                [8, 10, 91],
                [9, 10, 110],
            ]
        )

        self.test_data_numpy_1d = input_array_np_1d
        self.test_data_numpy_2d = input_array_np_2d
        self.test_data_numpy_3d = input_array_np_3d
        self.test_data_numpy_3d_constant = input_array_np_3d_constant

        self.test_data_pandas_1d = input_array_pd_1d
        self.test_data_pandas_2d = input_array_pd_2d
        self.test_data_pandas_3d = input_array_pd_3d
        self.test_data_pandas_3d_constant = input_array_pd_3d_constant

    @pytest.mark.unit
    def test_data_scaling_minmax_01(self):
        # 1D
        # Sample data generated from between 0 and 9
        input_array = self.test_data_numpy_1d
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        expected_output_3 = np.array([[9]])
        expected_output_2 = np.array([[0]])
        expected_output_1 = (input_array - expected_output_2) / (
            expected_output_3 - expected_output_2
        )
        np.testing.assert_array_equal(output_3, expected_output_3)
        np.testing.assert_array_equal(output_2, expected_output_2)
        np.testing.assert_array_equal(output_1, expected_output_1)

    @pytest.mark.unit
    def test_data_scaling_minmax_02(self):
        # 2D
        # Sample data generated from the expression (x_1 + 1)^2 between 0 and 9
        input_array = self.test_data_numpy_2d
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        expected_output_3 = np.array([[9, 100]])
        expected_output_2 = np.array([[0, 1]])
        expected_output_1 = (input_array - expected_output_2) / (
            expected_output_3 - expected_output_2
        )
        np.testing.assert_array_equal(output_3, expected_output_3)
        np.testing.assert_array_equal(output_2, expected_output_2)
        np.testing.assert_array_equal(output_1, expected_output_1)

    @pytest.mark.unit
    def test_data_scaling_minmax_03(self):
        # 3D
        # Sample data generated from the expression (x_1 + 1)^2 + x_2, x1: between 0 and 9, x2: between 10 and 19
        input_array = self.test_data_numpy_3d
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        expected_output_3 = np.array([[9, 19, 119]])
        expected_output_2 = np.array([[0, 10, 11]])
        expected_output_1 = (input_array - expected_output_2) / (
            expected_output_3 - expected_output_2
        )
        np.testing.assert_array_equal(output_3, expected_output_3)
        np.testing.assert_array_equal(output_2, expected_output_2)
        np.testing.assert_array_equal(output_1, expected_output_1)

    @pytest.mark.unit
    def test_data_scaling_minmax_04(self):
        # 3D with constant
        # Sample data generated from the expression (x_1 + 1)^2 + 10, x1: between 0 and 9
        input_array = self.test_data_numpy_3d_constant
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        expected_output_3 = np.array([[9, 10, 110]])
        expected_output_2 = np.array([[0, 10, 11]])
        scale = expected_output_3 - expected_output_2
        scale[scale == 0.0] = 1.0
        expected_output_1 = (input_array - expected_output_2) / scale
        np.testing.assert_array_equal(output_3, expected_output_3)
        np.testing.assert_array_equal(output_2, expected_output_2)
        np.testing.assert_array_equal(output_1, expected_output_1)

    @pytest.mark.unit
    def test_data_scaling_minmax_05(self):
        # TypeError with list
        input_array = self.test_data_numpy_2d.tolist()
        with pytest.raises(TypeError):
            FeatureScaling.data_scaling_minmax(input_array)

    @pytest.mark.unit
    def test_data_scaling_minmax_06(self):
        # 1D
        # Sample data generated from between 0 and 9
        input_array = self.test_data_pandas_1d
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        expected_output_3 = np.array([[9]])
        expected_output_2 = np.array([[0]])
        expected_output_1 = (input_array - expected_output_2) / (
            expected_output_3 - expected_output_2
        )
        np.testing.assert_array_equal(output_3, expected_output_3)
        np.testing.assert_array_equal(output_2, expected_output_2)
        np.testing.assert_array_equal(output_1, expected_output_1)

    @pytest.mark.unit
    def test_data_scaling_minmax_07(self):
        # 2D
        # Sample data generated from the expression (x_1 + 1)^2 between 0 and 9
        input_array = self.test_data_pandas_2d
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        expected_output_3 = np.array([[9, 100]])
        expected_output_2 = np.array([[0, 1]])
        expected_output_1 = (input_array - expected_output_2) / (
            expected_output_3 - expected_output_2
        )
        np.testing.assert_array_equal(output_3, expected_output_3)
        np.testing.assert_array_equal(output_2, expected_output_2)
        np.testing.assert_array_equal(output_1, expected_output_1)

    @pytest.mark.unit
    def test_data_scaling_minmax_08(self):
        # 3D
        # Sample data generated from the expression (x_1 + 1)^2 + x_2, x1: between 0 and 9, x2: between 10 and 19
        input_array = self.test_data_pandas_3d
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        expected_output_3 = np.array([[9, 19, 119]])
        expected_output_2 = np.array([[0, 10, 11]])
        expected_output_1 = (input_array - expected_output_2) / (
            expected_output_3 - expected_output_2
        )
        np.testing.assert_array_equal(output_3, expected_output_3)
        np.testing.assert_array_equal(output_2, expected_output_2)
        np.testing.assert_array_equal(output_1, expected_output_1)

    @pytest.mark.unit
    def test_data_scaling_minmax_09(self):
        # 3D with constant
        # Sample data generated from the expression (x_1 + 1)^2 + 10, x1: between 0 and 9
        input_array = self.test_data_pandas_3d_constant
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        expected_output_3 = np.array([[9, 10, 110]])
        expected_output_2 = np.array([[0, 10, 11]])
        scale = expected_output_3 - expected_output_2
        scale[scale == 0.0] = 1.0
        expected_output_1 = (input_array - expected_output_2) / scale
        np.testing.assert_array_equal(output_3, expected_output_3)
        np.testing.assert_array_equal(output_2, expected_output_2)
        np.testing.assert_array_equal(output_1, expected_output_1)

    @pytest.mark.unit
    def test_data_unscaling_minmax_01(self):
        # 1D
        # Sample data generated from between 0 and 9
        input_array = self.test_data_numpy_1d
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        un_output_1 = FeatureScaling.data_unscaling_minmax(output_1, output_2, output_3)
        np.testing.assert_array_equal(un_output_1, input_array)

    @pytest.mark.unit
    def test_data_unscaling_minmax_02(self):
        # 2D
        # Sample data generated from the expression (x_1 + 1)^2 between 0 and 9
        input_array = self.test_data_numpy_2d
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        un_output_1 = FeatureScaling.data_unscaling_minmax(output_1, output_2, output_3)
        np.testing.assert_array_equal(un_output_1, input_array)

    @pytest.mark.unit
    def test_data_unscaling_minmax_03(self):
        # 3D
        # Sample data generated from the expression (x_1 + 1)^2 + x_2, x1: between 0 and 9, x2: between 10 and 19
        input_array = self.test_data_numpy_3d
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        un_output_1 = FeatureScaling.data_unscaling_minmax(output_1, output_2, output_3)
        np.testing.assert_array_equal(un_output_1, input_array)

    @pytest.mark.unit
    def test_data_unscaling_minmax_04(self):
        # 3D with constant
        # Sample data generated from the expression (x_1 + 1)^2 + 10, x1: between 0 and 9
        input_array = self.test_data_numpy_3d_constant
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        un_output_1 = FeatureScaling.data_unscaling_minmax(output_1, output_2, output_3)
        np.testing.assert_array_equal(un_output_1, input_array)

    @pytest.mark.unit
    def test_data_unscaling_minmax_05(self):
        # 2D
        # Sample data generated from the expression (x_1 + 1)^2 between 0 and 9
        input_array = self.test_data_numpy_2d
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)

        min_array = np.array([[1]])
        max_array = np.array([[5]])
        with pytest.raises(IndexError):
            FeatureScaling.data_unscaling_minmax(output_1, min_array, max_array)

    @pytest.mark.unit
    def test_data_unscaling_minmax_06(self):
        # 2D
        # Sample data generated from the expression (x_1 + 1)^2 between 0 and 9
        input_array = self.test_data_numpy_2d
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)

        min_array = np.array([[1, 2, 3]])
        max_array = np.array([[5, 6, 7]])
        with pytest.raises(IndexError):
            FeatureScaling.data_unscaling_minmax(output_1, min_array, max_array)


class SamplingMethodsTestCases(unittest.TestCase):
    def setUp(self):
        input_array_np_1d = np.array([[0], [1], [2], [3], [4], [5], [6], [7], [8], [9]])
        input_array_np_2d = np.array(
            [
                [0, 1],
                [1, 4],
                [2, 9],
                [3, 16],
                [4, 25],
                [5, 36],
                [6, 49],
                [7, 64],
                [8, 81],
                [9, 100],
            ]
        )
        input_array_np_3d = np.array(
            [
                [0, 10, 11],
                [1, 11, 15],
                [2, 12, 21],
                [3, 13, 29],
                [4, 14, 39],
                [5, 15, 51],
                [6, 16, 65],
                [7, 17, 81],
                [8, 18, 99],
                [9, 19, 119],
            ]
        )
        input_array_pd = pd.DataFrame(
            {
                "x1": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                "x2": [10, 10, 10, 10, 10, 10, 10, 10, 10, 10],
                "y1": [11, 14, 19, 26, 35, 46, 59, 74, 91, 110],
                "y2": [21, 24, 29, 36, 45, 56, 69, 84, 101, 120],
            }
        )

        self.test_data_numpy_1d = input_array_np_1d
        self.test_data_numpy_2d = input_array_np_2d
        self.test_data_numpy_3d = input_array_np_3d
        self.test_data_pandas = input_array_pd

    """
   
    test_nearest_neighbour_01: Test behaviour with array N x d = (10,3) data and a=[-0.5,1]
    test_nearest_neighbour_02: Test behaviour with array N x d = (10,2) data and a=[-0.5]
    test_nearest_neighbour_03: Test behaviour with array N x d = (10,1) data and a=[]
    test_nearest_neighbour_04: working even with different input size
    test_nearest_neighbour_05: Test behaviour raise ValueError if dimension of point is not matching with array N x d = (10,3) 
    
    test_points_selection_01: Test behaviour with array N x d = (10,3) data and a=[[-0.5,10],[10,100]]
    test_points_selection_02: Test behaviour with array N x d = (10,2) data and a=[[-0.5],[10]]
    test_points_selection_03: Test behaviour with array N x d = (10,1) data and a=[[],[]] <---both return the first element
    test_points_selection_04: Test behaviour raise ValueError if dimension of point is not matching with array N x d = (10,3); small
    test_points_selection_05: Test behaviour raise ValueError if dimension of point is not matching with array N x d = (10,3); large
    
    test_sample_point_selection_01: selection - Test behaviour with array N x d = (10,3) data and a=[[0,0],[10,19]]
    test_sample_point_selection_02: selection - Test behaviour with array N x d = (10,2) data and a=[[0],[7]]
    test_sample_point_selection_03: selection - Test behaviour with array N x d = (10,1) data and a=[[],[]] <---both return the first element, so return only 1 []
    test_sample_point_selection_04: selection - Test behaviour raise ValueError if dimension of point is not matching with array N x d = (10,3); large
    test_sample_point_selection_05: selection - Test behaviour raise ValueError if dimension of point is not matching with array N x d = (10,3); large
    
    test_points_selection_06: creation - Test behaviour with points dimension should be d with array N x d = (10,3) 
    test_points_selection_07: creation - Test behaviour with points dimension should be d with array N x d = (10,2) 
    test_points_selection_08: creation - Test behaviour with points dimension should be d with array N x d = (10,1) 
    test_points_selection_09: creation - raise IndexError if dimension of point is not matching with array N x d = (10,3); small
    test_points_selection_10: creation - raise IndexError if dimension of point is not matching with array N x d = (10,3); large
    
    test_points_selection_01: Test behaviour with n = 3
    test_points_selection_02: Test behaviour with n = 1
    test_points_selection_03: Test behaviour with n = 0
    test_points_selection_04: Test behaviour with n = -1
    test_points_selection_04: Test behaviour with n = 2.9
    
    test_base_conversion_01: Test behaviour with 5 to base 2
    test_base_conversion_02: Test behaviour with 57 to base 47
    test_base_conversion_03: Test behaviour with negative base - works, returns always 0
    test_base_conversion_04: Test behaviour raise ZeroDivisionError with 0 base
    test_base_conversion_05: Test behaviour with 1 base, infinity loop
    
    test_prime_base_to_decimal_01: Test behaviour with 0.01 in base 2 to base 10
    test_prime_base_to_decimal_02: Test behaviour with 0.01 in base 20 to base 10
    test_prime_base_to_decimal_03: working with base 1
    test_prime_base_to_decimal_04: working with base -1

    test_selection_columns_preprocessing_01 - 11: Test behaviour of function for a dataframe input under a variety of xlabels and ylabels conditions. 
        See test docstrings for more specific details.
    test_selection_columns_preprocessing_14 - 26: Repeats above tests 1-13 for case where input is a numpy array rather than a dataframe
    """

    def _create_sampling(self, input_array, sample_points):
        sampling = SamplingMethods()
        sampling.data_headers = [i for i in range(0, sample_points.shape[1] + 1)]
        sampling.x_data = np.zeros((2, input_array.shape[1] - 1))
        return sampling

    @pytest.mark.unit
    def test_nearest_neighbour_01(self):
        input_array = self.test_data_numpy_3d
        self.x_data = np.zeros((2, input_array.shape[1] - 1))
        closest_point = SamplingMethods.nearest_neighbour(self, input_array, [-0.5, 1])
        np.testing.assert_array_equal(closest_point, input_array[0, :])

    @pytest.mark.unit
    def test_nearest_neighbour_02(self):
        input_array = self.test_data_numpy_2d
        self.x_data = np.zeros((2, input_array.shape[1] - 1))
        closest_point = SamplingMethods.nearest_neighbour(self, input_array, [-0.5])
        np.testing.assert_array_equal(closest_point, input_array[0, :])

    @pytest.mark.unit
    def test_nearest_neighbour_03(self):
        input_array = self.test_data_numpy_1d
        self.x_data = np.zeros((2, input_array.shape[1] - 1))
        closest_point = SamplingMethods.nearest_neighbour(self, input_array, [])
        np.testing.assert_array_equal(closest_point, input_array[0, :])

    @pytest.mark.unit
    def test_nearest_neighbour_04(self):
        input_array = self.test_data_numpy_3d
        self.x_data = np.zeros((2, input_array.shape[1] - 1))
        closest_point = SamplingMethods.nearest_neighbour(self, input_array, [0.5])

    @pytest.mark.unit
    def test_nearest_neighbour_05(self):
        input_array = self.test_data_numpy_3d
        self.x_data = np.zeros((2, input_array.shape[1] - 1))
        with pytest.raises(ValueError):
            closest_point = SamplingMethods.nearest_neighbour(
                self, input_array, [0.5, 0.9, 10]
            )

    @pytest.mark.unit
    def test_points_selection_01(self):
        input_array = self.test_data_numpy_3d
        generated_sample_points = np.array([[-0.5, 10], [10, 100]])
        sampling_methods = self._create_sampling(input_array, generated_sample_points)
        equivalent_points = sampling_methods.points_selection(
            input_array, generated_sample_points
        )
        np.testing.assert_array_equal(equivalent_points[0], input_array[0, :])
        np.testing.assert_array_equal(equivalent_points[1], input_array[-1, :])

    @pytest.mark.unit
    def test_points_selection_02(self):
        input_array = self.test_data_numpy_2d
        generated_sample_points = np.array([[-0.5], [10]])
        sampling_methods = self._create_sampling(input_array, generated_sample_points)
        equivalent_points = sampling_methods.points_selection(
            input_array, generated_sample_points
        )
        np.testing.assert_array_equal(equivalent_points[0], input_array[0, :])
        np.testing.assert_array_equal(equivalent_points[1], input_array[-1, :])

    @pytest.mark.unit
    def test_points_selection_03(self):
        input_array = self.test_data_numpy_1d
        generated_sample_points = np.array([[], []])
        sampling_methods = self._create_sampling(input_array, generated_sample_points)
        equivalent_points = sampling_methods.points_selection(
            input_array, generated_sample_points
        )
        np.testing.assert_array_equal(equivalent_points[0], input_array[0, :])
        np.testing.assert_array_equal(equivalent_points[1], input_array[0, :])

    @pytest.mark.unit
    def test_points_selection_04(self):
        input_array = self.test_data_numpy_3d
        generated_sample_points = np.array([[0.5], [10]])
        sampling_methods = self._create_sampling(input_array, generated_sample_points)
        with pytest.raises(ValueError):
            equivalent_points = sampling_methods.points_selection(
                input_array, generated_sample_points
            )

    @pytest.mark.unit
    def test_points_selection_05(self):
        input_array = self.test_data_numpy_3d
        generated_sample_points = np.array([[0.5, 0.7, 10], [10, 0.9, 20]])
        sampling_methods = self._create_sampling(input_array, generated_sample_points)
        with pytest.raises(ValueError):
            equivalent_points = sampling_methods.points_selection(
                input_array, generated_sample_points
            )

    @pytest.mark.unit
    def test_sample_point_selection_01(self):
        input_array = self.test_data_numpy_3d
        generated_sample_points = np.array([[0, 0], [10, 19]])
        sampling_methods = self._create_sampling(input_array, generated_sample_points)
        unique_sample_points = sampling_methods.sample_point_selection(
            input_array, generated_sample_points, sampling_type="selection"
        )
        np.testing.assert_array_equal(unique_sample_points[0], input_array[0, :])
        np.testing.assert_array_equal(unique_sample_points[1], input_array[-1, :])

    @pytest.mark.unit
    def test_sample_point_selection_02(self):
        input_array = self.test_data_numpy_2d
        generated_sample_points = np.array([[0], [7]])
        sampling_methods = self._create_sampling(input_array, generated_sample_points)
        unique_sample_points = sampling_methods.sample_point_selection(
            input_array, generated_sample_points, sampling_type="selection"
        )
        np.testing.assert_array_equal(unique_sample_points[0], input_array[0, :])
        np.testing.assert_array_equal(unique_sample_points[1], input_array[-1, :])

    @pytest.mark.unit
    def test_sample_point_selection_03(self):
        input_array = self.test_data_numpy_1d
        generated_sample_points = np.array([[], []])
        sampling_methods = self._create_sampling(input_array, generated_sample_points)
        unique_sample_points = sampling_methods.sample_point_selection(
            input_array, generated_sample_points, sampling_type="selection"
        )
        np.testing.assert_array_equal(unique_sample_points[0], input_array[0, :])

    @pytest.mark.unit
    def test_sample_point_selection_04(self):
        input_array = self.test_data_numpy_3d
        generated_sample_points = np.array([[0.5], [7]])
        sampling_methods = self._create_sampling(input_array, generated_sample_points)
        with pytest.raises(ValueError):
            unique_sample_points = sampling_methods.sample_point_selection(
                input_array, generated_sample_points, sampling_type="selection"
            )

    @pytest.mark.unit
    def test_sample_point_selection_05(self):
        input_array = self.test_data_numpy_3d
        generated_sample_points = np.array([[0.5, 1, 10], [7, 19, 20]])
        sampling_methods = self._create_sampling(input_array, generated_sample_points)
        with pytest.raises(ValueError):
            unique_sample_points = sampling_methods.sample_point_selection(
                input_array, generated_sample_points, sampling_type="selection"
            )

    @pytest.mark.unit
    def test_sample_point_selection_06(self):
        input_array = self.test_data_numpy_3d
        generated_sample_points = np.array([[0.5, 11, 3], [7, 19, 4]])
        sampling_methods = self._create_sampling(input_array, generated_sample_points)
        unique_sample_points = sampling_methods.sample_point_selection(
            input_array, generated_sample_points, sampling_type="creation"
        )
        min_, max_ = input_array[0, :], input_array[1, :]
        testing = min_ + generated_sample_points * (max_ - min_)
        np.testing.assert_array_equal(testing, unique_sample_points)

    @pytest.mark.unit
    def test_sample_point_selection_07(self):
        input_array = self.test_data_numpy_2d
        generated_sample_points = np.array([[0.5, 1], [7, 19]])
        sampling_methods = self._create_sampling(input_array, generated_sample_points)
        unique_sample_points = sampling_methods.sample_point_selection(
            input_array, generated_sample_points, sampling_type="creation"
        )
        min_, max_ = input_array[0, :], input_array[1, :]
        testing = min_ + generated_sample_points * (max_ - min_)
        np.testing.assert_array_equal(testing, unique_sample_points)

    @pytest.mark.unit
    def test_sample_point_selection_08(self):
        input_array = self.test_data_numpy_1d
        generated_sample_points = np.array([[0.5], [7]])
        sampling_methods = self._create_sampling(input_array, generated_sample_points)
        sampling_methods.x_data = np.zeros((2, input_array.shape[1] - 1))
        unique_sample_points = sampling_methods.sample_point_selection(
            input_array, generated_sample_points, sampling_type="creation"
        )
        min_, max_ = input_array[0, :], input_array[1, :]
        testing = min_ + generated_sample_points * (max_ - min_)
        np.testing.assert_array_equal(testing, unique_sample_points)

    @pytest.mark.unit
    def test_sample_point_selection_09(self):
        input_array = self.test_data_numpy_3d
        generated_sample_points = np.array([[], []])
        sampling_methods = self._create_sampling(input_array, generated_sample_points)
        with pytest.raises(IndexError):
            unique_sample_points = sampling_methods.sample_point_selection(
                input_array, generated_sample_points, sampling_type="creation"
            )

    @pytest.mark.unit
    def test_sample_point_selection_10(self):
        input_array = self.test_data_numpy_3d
        generated_sample_points = np.array([[0.5, 1, 10, 11], [7, 19, 10, 12]])
        sampling_methods = self._create_sampling(input_array, generated_sample_points)
        with pytest.raises(IndexError):
            unique_sample_points = sampling_methods.sample_point_selection(
                input_array, generated_sample_points, sampling_type="creation"
            )

    @pytest.mark.unit
    def test_prime_number_generator_01(self):
        prime_list = SamplingMethods.prime_number_generator(self, 3)
        np.testing.assert_array_equal(prime_list, [2, 3, 5])

    @pytest.mark.unit
    def test_prime_number_generator_02(self):
        prime_list = SamplingMethods.prime_number_generator(self, 1)
        np.testing.assert_array_equal(prime_list, [2])

    @pytest.mark.unit
    def test_prime_number_generator_03(self):
        prime_list = SamplingMethods.prime_number_generator(self, 0)
        np.testing.assert_array_equal(prime_list, [])

    @pytest.mark.unit
    def test_prime_number_generator_04(self):
        prime_list = SamplingMethods.prime_number_generator(self, -1)
        np.testing.assert_array_equal(prime_list, [])

    @pytest.mark.unit
    def test_prime_number_generator_05(self):
        prime_list = SamplingMethods.prime_number_generator(self, 2.9)
        np.testing.assert_array_equal(prime_list, [2, 3, 5])

    @pytest.mark.unit
    def test_base_conversion_01(self):
        string_representation = SamplingMethods.base_conversion(self, 5, 2)
        assert string_representation == ["1", "0", "1"]

    @pytest.mark.unit
    def test_base_conversion_02(self):
        string_representation = SamplingMethods.base_conversion(self, 57, 47)
        assert string_representation == ["1", "10"]

    @pytest.mark.unit
    def test_base_conversion_03(self):
        string_representation = SamplingMethods.base_conversion(self, 10, -1)
        assert string_representation == ["0"]

    @pytest.mark.unit
    def test_base_conversion_04(self):
        with pytest.raises(ZeroDivisionError):
            string_representation = SamplingMethods.base_conversion(self, 10, 0)

    # @pytest.mark.unit
    # def test_base_conversion_05(self):
    #     string_representation = SamplingMethods.base_conversion(self, 10, 1)

    @pytest.mark.unit
    def test_prime_base_to_decimal_01(self):
        string_representation = SamplingMethods.prime_base_to_decimal(
            self, ["0", "0", "1"], 2
        )
        assert 0.25 == string_representation

    @pytest.mark.unit
    def test_prime_base_to_decimal_02(self):
        string_representation = SamplingMethods.prime_base_to_decimal(
            self, ["0", "0", "1"], 20
        )
        assert 0.0025 == string_representation

    @pytest.mark.unit
    def test_prime_base_to_decimal_03(self):
        string_representation = SamplingMethods.prime_base_to_decimal(
            self, ["0", "0", "1"], 1
        )

    @pytest.mark.unit
    def test_prime_base_to_decimal_04(self):
        string_representation = SamplingMethods.prime_base_to_decimal(
            self, ["0", "0", "1"], -1
        )

    @pytest.mark.unit
    def test_data_sequencing(self):
        sampling_methods = SamplingMethods()
        sequence_decimal = sampling_methods.data_sequencing(3, 2)

    @pytest.mark.unit
    def test_selection_columns_preprocessing_01(self):
        """
        Test behaviour with no labels specified
        """
        sampling_methods = SamplingMethods()
        sampling_methods.selection_columns_preprocessing(
            self.test_data_pandas, None, None
        )
        assert sampling_methods.df_flag == True
        assert sampling_methods.data.shape[1] == self.test_data_pandas.shape[1]
        assert sampling_methods.x_data.shape[1] == self.test_data_pandas.shape[1] - 1
        assert (
            sampling_methods.data_headers
            == self.test_data_pandas.columns.values.tolist()
        )
        assert (
            sampling_methods.data_headers_xvars
            == self.test_data_pandas.columns.values.tolist()[:-1]
        )
        np.testing.assert_array_equal(
            sampling_methods.data, self.test_data_pandas.values
        )
        np.testing.assert_array_equal(
            sampling_methods.x_data, self.test_data_pandas.values[:, :-1]
        )

    @pytest.mark.unit
    def test_selection_columns_preprocessing_02(self):
        """
        Test behaviour with only xlabels specified with multiple elements, same order as in dataframe
        """
        sampling_methods = SamplingMethods()
        x_lab = ["x1", "x2"]
        sampling_methods.selection_columns_preprocessing(
            self.test_data_pandas, x_lab, None
        )
        assert sampling_methods.df_flag == True
        assert sampling_methods.data.shape[1] == self.test_data_pandas.shape[1]
        assert sampling_methods.x_data.shape[1] == len(x_lab)
        assert (
            sampling_methods.data_headers
            == self.test_data_pandas.columns.values.tolist()
        )
        assert sampling_methods.data_headers_xvars == x_lab
        np.testing.assert_array_equal(
            sampling_methods.data, self.test_data_pandas.values
        )
        np.testing.assert_array_equal(
            sampling_methods.x_data, self.test_data_pandas.values[:, :-2]
        )

    @pytest.mark.unit
    def test_selection_columns_preprocessing_03(self):
        """
        Test behaviour with only xlabels specified with multiple elements, different order from dataframe
        """
        sampling_methods = SamplingMethods()
        x_lab = ["y1", "x1"]
        sampling_methods.selection_columns_preprocessing(
            self.test_data_pandas, x_lab, None
        )
        assert sampling_methods.df_flag == True
        assert sampling_methods.data.shape[1] == self.test_data_pandas.shape[1]
        assert sampling_methods.x_data.shape[1] == len(x_lab)
        assert sampling_methods.data_headers == ["y1", "x1", "x2", "y2"]
        assert sampling_methods.data_headers_xvars == x_lab
        np.testing.assert_array_equal(
            sampling_methods.data,
            self.test_data_pandas[["y1", "x1", "x2", "y2"]].values,
        )
        np.testing.assert_array_equal(
            sampling_methods.x_data, self.test_data_pandas[x_lab].values
        )

    @pytest.mark.unit
    def test_selection_columns_preprocessing_04(self):
        """
        Test behaviour with only xlabels specified with only 1 element
        """
        sampling_methods = SamplingMethods()
        x_lab = ["y2"]
        expected_order = ["y2", "x1", "x2", "y1"]
        sampling_methods.selection_columns_preprocessing(
            self.test_data_pandas, x_lab, None
        )
        assert sampling_methods.df_flag == True
        assert sampling_methods.data.shape[1] == self.test_data_pandas.shape[1]
        assert sampling_methods.x_data.shape[1] == len(x_lab)
        assert sampling_methods.data_headers == expected_order
        assert sampling_methods.data_headers_xvars == x_lab
        np.testing.assert_array_equal(
            sampling_methods.data, self.test_data_pandas[expected_order].values
        )
        np.testing.assert_array_equal(
            sampling_methods.x_data, self.test_data_pandas[x_lab].values
        )

    @pytest.mark.unit
    def test_selection_columns_preprocessing_05(self):
        """
        Test behaviour with only ylabels specified with multiple elements, same order as in dataframe
        """
        sampling_methods = SamplingMethods()
        y_lab = ["y1", "y2"]
        expected_order = ["x1", "x2", "y1", "y2"]
        expected_x_lab = ["x1", "x2"]
        sampling_methods.selection_columns_preprocessing(
            self.test_data_pandas, None, y_lab
        )
        assert sampling_methods.df_flag == True
        assert sampling_methods.data.shape[1] == self.test_data_pandas.shape[1]
        assert sampling_methods.x_data.shape[1] == self.test_data_pandas.shape[1] - len(
            y_lab
        )
        assert sampling_methods.data_headers == expected_order
        assert sampling_methods.data_headers_xvars == expected_x_lab
        np.testing.assert_array_equal(
            sampling_methods.data, self.test_data_pandas[expected_order].values
        )
        np.testing.assert_array_equal(
            sampling_methods.x_data, self.test_data_pandas[expected_x_lab].values
        )

    @pytest.mark.unit
    def test_selection_columns_preprocessing_06(self):
        """
        Test behaviour with only ylabels specified with multiple elements, different order from dataframe
        """
        sampling_methods = SamplingMethods()
        y_lab = ["y1", "x1"]
        expected_order = ["x2", "y2", "y1", "x1"]
        expected_x_lab = ["x2", "y2"]
        sampling_methods.selection_columns_preprocessing(
            self.test_data_pandas, None, y_lab
        )
        assert sampling_methods.df_flag == True
        assert sampling_methods.data.shape[1] == self.test_data_pandas.shape[1]
        assert sampling_methods.x_data.shape[1] == self.test_data_pandas.shape[1] - len(
            y_lab
        )
        assert sampling_methods.data_headers == expected_order
        assert sampling_methods.data_headers_xvars == expected_x_lab
        np.testing.assert_array_equal(
            sampling_methods.data, self.test_data_pandas[expected_order].values
        )
        np.testing.assert_array_equal(
            sampling_methods.x_data, self.test_data_pandas[expected_x_lab].values
        )

    @pytest.mark.unit
    def test_selection_columns_preprocessing_07(self):
        """
        Test behaviour with only ylabels specified with only 1 element
        """
        sampling_methods = SamplingMethods()
        y_lab = ["x2"]
        expected_order = ["x1", "y1", "y2", "x2"]
        expected_x_lab = ["x1", "y1", "y2"]
        sampling_methods.selection_columns_preprocessing(
            self.test_data_pandas, None, y_lab
        )
        assert sampling_methods.df_flag == True
        assert sampling_methods.data.shape[1] == self.test_data_pandas.shape[1]
        assert sampling_methods.x_data.shape[1] == self.test_data_pandas.shape[1] - len(
            y_lab
        )
        assert sampling_methods.data_headers == expected_order
        assert sampling_methods.data_headers_xvars == expected_x_lab
        np.testing.assert_array_equal(
            sampling_methods.data, self.test_data_pandas[expected_order].values
        )
        np.testing.assert_array_equal(
            sampling_methods.x_data, self.test_data_pandas[expected_x_lab].values
        )

    @pytest.mark.unit
    def test_selection_columns_preprocessing_11(self):
        """
        Test behaviour with both xlabels and ylabels specified; multiple x and single y
        """
        sampling_methods = SamplingMethods()
        x_lab = ["x1", "x2"]
        y_lab = ["y2"]
        expected_order = x_lab + y_lab
        sampling_methods.selection_columns_preprocessing(
            self.test_data_pandas, x_lab, y_lab
        )
        assert sampling_methods.df_flag == True
        assert sampling_methods.data.shape[1] == len(expected_order)
        assert sampling_methods.x_data.shape[1] == len(x_lab)
        assert sampling_methods.data_headers == expected_order
        assert sampling_methods.data_headers_xvars == x_lab
        np.testing.assert_array_equal(
            sampling_methods.data, self.test_data_pandas[expected_order].values
        )
        np.testing.assert_array_equal(
            sampling_methods.x_data, self.test_data_pandas[x_lab].values
        )

    @pytest.mark.unit
    def test_selection_columns_preprocessing_08(self):
        """
        Test behaviour with both xlabels and ylabels specified; single x and single y; ordered
        """
        sampling_methods = SamplingMethods()
        x_lab = ["x1"]
        y_lab = ["y1"]
        expected_order = x_lab + y_lab
        sampling_methods.selection_columns_preprocessing(
            self.test_data_pandas, x_lab, y_lab
        )
        assert sampling_methods.df_flag == True
        assert sampling_methods.data.shape[1] == len(x_lab) + len(y_lab)
        assert sampling_methods.x_data.shape[1] == len(x_lab)
        assert sampling_methods.data_headers == expected_order
        assert sampling_methods.data_headers_xvars == x_lab
        np.testing.assert_array_equal(
            sampling_methods.data, self.test_data_pandas[expected_order].values
        )
        np.testing.assert_array_equal(
            sampling_methods.x_data, self.test_data_pandas[x_lab].values
        )

    @pytest.mark.unit
    def test_selection_columns_preprocessing_09(self):
        """
        Test behaviour with both xlabels and ylabels specified; single x and multiple y
        """
        sampling_methods = SamplingMethods()
        x_lab = ["x1"]
        y_lab = ["y1", "y2"]
        expected_order = x_lab + y_lab
        sampling_methods.selection_columns_preprocessing(
            self.test_data_pandas, x_lab, y_lab
        )
        assert sampling_methods.df_flag == True
        assert sampling_methods.data.shape[1] == len(expected_order)
        assert sampling_methods.x_data.shape[1] == len(x_lab)
        assert sampling_methods.data_headers == expected_order
        assert sampling_methods.data_headers_xvars == x_lab
        np.testing.assert_array_equal(
            sampling_methods.data, self.test_data_pandas[expected_order].values
        )
        np.testing.assert_array_equal(
            sampling_methods.x_data, self.test_data_pandas[x_lab].values
        )

    @pytest.mark.unit
    def test_selection_columns_preprocessing_10(self):
        """
        Test behaviour with both xlabels and ylabels specified; single x and single y; unordered
        """
        sampling_methods = SamplingMethods()
        x_lab = ["y1"]
        y_lab = ["x2"]
        expected_order = x_lab + y_lab
        sampling_methods.selection_columns_preprocessing(
            self.test_data_pandas, x_lab, y_lab
        )
        assert sampling_methods.df_flag == True
        assert sampling_methods.data.shape[1] == len(expected_order)
        assert sampling_methods.x_data.shape[1] == len(x_lab)
        assert sampling_methods.data_headers == expected_order
        assert sampling_methods.data_headers_xvars == x_lab
        np.testing.assert_array_equal(
            sampling_methods.data, self.test_data_pandas[expected_order].values
        )
        np.testing.assert_array_equal(
            sampling_methods.x_data, self.test_data_pandas[x_lab].values
        )

    @pytest.mark.unit
    def test_selection_columns_preprocessing_12(self):
        """
        Test behaviour with both xlabels and ylabels specified; multiple x and y, all initial columns present
        """
        sampling_methods = SamplingMethods()
        x_lab = ["y1", "x2"]
        y_lab = ["x1", "y2"]
        expected_order = x_lab + y_lab
        sampling_methods.selection_columns_preprocessing(
            self.test_data_pandas, x_lab, y_lab
        )
        assert sampling_methods.df_flag == True
        assert sampling_methods.data.shape[1] == len(expected_order)
        assert sampling_methods.x_data.shape[1] == len(x_lab)
        assert sampling_methods.data_headers == expected_order
        assert sampling_methods.data_headers_xvars == x_lab
        np.testing.assert_array_equal(
            sampling_methods.data, self.test_data_pandas[expected_order].values
        )
        np.testing.assert_array_equal(
            sampling_methods.x_data, self.test_data_pandas[x_lab].values
        )

    @pytest.mark.unit
    def test_selection_columns_preprocessing_13(self):
        """
        Test behaviour when non-existent column name is supplied in label - should raise ValueError
        """
        sampling_methods = SamplingMethods()
        x_lab = ["x1"]
        y_lab = ["y3"]
        expected_order = x_lab + y_lab

        with pytest.raises(IndexError):
            sampling_methods.selection_columns_preprocessing(
                self.test_data_pandas, x_lab, y_lab
            )

    @pytest.mark.unit
    def test_selection_columns_preprocessing_14(self):
        sampling_methods = SamplingMethods()
        sampling_methods.selection_columns_preprocessing(
            self.test_data_pandas.values, None, None
        )
        assert sampling_methods.df_flag == False
        assert sampling_methods.data.shape[1] == self.test_data_pandas.shape[1]
        assert sampling_methods.x_data.shape[1] == self.test_data_pandas.shape[1] - 1
        assert sampling_methods.data_headers == [
            i for i in range(0, self.test_data_pandas.shape[1])
        ]
        assert sampling_methods.data_headers_xvars == [
            i for i in range(0, self.test_data_pandas.shape[1] - 1)
        ]
        np.testing.assert_array_equal(
            sampling_methods.data, self.test_data_pandas.values
        )
        np.testing.assert_array_equal(
            sampling_methods.x_data, self.test_data_pandas.values[:, :-1]
        )

    @pytest.mark.unit
    def test_selection_columns_preprocessing_15(self):
        sampling_methods = SamplingMethods()
        x_lab = [0, 1]
        expected_order = [0, 1, 2, 3]
        sampling_methods.selection_columns_preprocessing(
            self.test_data_pandas.values, x_lab, None
        )
        assert sampling_methods.df_flag == False
        assert sampling_methods.data.shape[1] == self.test_data_pandas.shape[1]
        assert sampling_methods.x_data.shape[1] == len(x_lab)
        assert sampling_methods.data_headers == expected_order
        assert sampling_methods.data_headers_xvars == x_lab
        np.testing.assert_array_equal(
            sampling_methods.data, self.test_data_pandas.values[:, expected_order]
        )
        np.testing.assert_array_equal(
            sampling_methods.x_data, self.test_data_pandas.values[:, x_lab]
        )

    @pytest.mark.unit
    def test_selection_columns_preprocessing_16(self):
        sampling_methods = SamplingMethods()
        x_lab = [2, 0]
        expected_order = [2, 0, 1, 3]
        sampling_methods.selection_columns_preprocessing(
            self.test_data_pandas.values, x_lab, None
        )
        assert sampling_methods.df_flag == False
        assert sampling_methods.data.shape[1] == self.test_data_pandas.shape[1]
        assert sampling_methods.x_data.shape[1] == len(x_lab)
        assert sampling_methods.data_headers == expected_order
        assert sampling_methods.data_headers_xvars == x_lab
        np.testing.assert_array_equal(
            sampling_methods.data, self.test_data_pandas.values[:, expected_order]
        )
        np.testing.assert_array_equal(
            sampling_methods.x_data, self.test_data_pandas.values[:, x_lab]
        )

    @pytest.mark.unit
    def test_selection_columns_preprocessing_17(self):
        sampling_methods = SamplingMethods()
        x_lab = [3]
        expected_order = [3, 0, 1, 2]
        sampling_methods.selection_columns_preprocessing(
            self.test_data_pandas.values, x_lab, None
        )
        assert sampling_methods.df_flag == False
        assert sampling_methods.data.shape[1] == self.test_data_pandas.shape[1]
        assert sampling_methods.x_data.shape[1] == len(x_lab)
        assert sampling_methods.data_headers == expected_order
        assert sampling_methods.data_headers_xvars == x_lab
        np.testing.assert_array_equal(
            sampling_methods.data, self.test_data_pandas.values[:, expected_order]
        )
        np.testing.assert_array_equal(
            sampling_methods.x_data, self.test_data_pandas.values[:, x_lab]
        )

    @pytest.mark.unit
    def test_selection_columns_preprocessing_18(self):
        sampling_methods = SamplingMethods()
        y_lab = [2, 3]
        expected_order = [0, 1, 2, 3]
        expected_x_lab = [0, 1]
        sampling_methods.selection_columns_preprocessing(
            self.test_data_pandas.values, None, y_lab
        )
        assert sampling_methods.df_flag == False
        assert sampling_methods.data.shape[1] == self.test_data_pandas.shape[1]
        assert sampling_methods.x_data.shape[1] == self.test_data_pandas.shape[1] - len(
            y_lab
        )
        assert sampling_methods.data_headers == expected_order
        assert sampling_methods.data_headers_xvars == expected_x_lab
        np.testing.assert_array_equal(
            sampling_methods.data, self.test_data_pandas.values[:, expected_order]
        )
        np.testing.assert_array_equal(
            sampling_methods.x_data, self.test_data_pandas.values[:, expected_x_lab]
        )

    @pytest.mark.unit
    def test_selection_columns_preprocessing_19(self):
        sampling_methods = SamplingMethods()
        y_lab = [2, 0]
        expected_order = [1, 3, 2, 0]
        expected_x_lab = [1, 3]
        sampling_methods.selection_columns_preprocessing(
            self.test_data_pandas.values, None, y_lab
        )
        assert sampling_methods.df_flag == False
        assert sampling_methods.data.shape[1] == self.test_data_pandas.shape[1]
        assert sampling_methods.x_data.shape[1] == self.test_data_pandas.shape[1] - len(
            y_lab
        )
        assert sampling_methods.data_headers == expected_order
        assert sampling_methods.data_headers_xvars == expected_x_lab
        np.testing.assert_array_equal(
            sampling_methods.data, self.test_data_pandas.values[:, expected_order]
        )
        np.testing.assert_array_equal(
            sampling_methods.x_data, self.test_data_pandas.values[:, expected_x_lab]
        )

    @pytest.mark.unit
    def test_selection_columns_preprocessing_20(self):
        sampling_methods = SamplingMethods()
        y_lab = [1]
        expected_order = [0, 2, 3, 1]
        expected_x_lab = [0, 2, 3]
        sampling_methods.selection_columns_preprocessing(
            self.test_data_pandas.values, None, y_lab
        )
        assert sampling_methods.df_flag == False
        assert sampling_methods.data.shape[1] == self.test_data_pandas.shape[1]
        assert sampling_methods.x_data.shape[1] == self.test_data_pandas.shape[1] - len(
            y_lab
        )
        assert sampling_methods.data_headers == expected_order
        assert sampling_methods.data_headers_xvars == expected_x_lab
        np.testing.assert_array_equal(
            sampling_methods.data, self.test_data_pandas.values[:, expected_order]
        )
        np.testing.assert_array_equal(
            sampling_methods.x_data, self.test_data_pandas.values[:, expected_x_lab]
        )

    @pytest.mark.unit
    def test_selection_columns_preprocessing_21(self):
        sampling_methods = SamplingMethods()
        x_lab = [0, 1]
        y_lab = [3]
        expected_order = x_lab + y_lab
        sampling_methods.selection_columns_preprocessing(
            self.test_data_pandas.values, x_lab, y_lab
        )
        assert sampling_methods.df_flag == False
        assert sampling_methods.data.shape[1] == len(expected_order)
        assert sampling_methods.x_data.shape[1] == len(x_lab)
        assert sampling_methods.data_headers == expected_order
        assert sampling_methods.data_headers_xvars == x_lab
        np.testing.assert_array_equal(
            sampling_methods.data, self.test_data_pandas.values[:, expected_order]
        )
        np.testing.assert_array_equal(
            sampling_methods.x_data, self.test_data_pandas.values[:, x_lab]
        )

    @pytest.mark.unit
    def test_selection_columns_preprocessing_22(self):
        sampling_methods = SamplingMethods()
        x_lab = [0]
        y_lab = [2]
        expected_order = x_lab + y_lab
        sampling_methods.selection_columns_preprocessing(
            self.test_data_pandas.values, x_lab, y_lab
        )
        assert sampling_methods.df_flag == False
        assert sampling_methods.data.shape[1] == len(x_lab) + len(y_lab)
        assert sampling_methods.x_data.shape[1] == len(x_lab)
        assert sampling_methods.data_headers == expected_order
        assert sampling_methods.data_headers_xvars == x_lab
        np.testing.assert_array_equal(
            sampling_methods.data, self.test_data_pandas.values[:, expected_order]
        )
        np.testing.assert_array_equal(
            sampling_methods.x_data, self.test_data_pandas.values[:, x_lab]
        )

    @pytest.mark.unit
    def test_selection_columns_preprocessing_23(self):
        sampling_methods = SamplingMethods()
        x_lab = [0]
        y_lab = [2, 3]
        expected_order = x_lab + y_lab
        sampling_methods.selection_columns_preprocessing(
            self.test_data_pandas.values, x_lab, y_lab
        )
        assert sampling_methods.df_flag == False
        assert sampling_methods.data.shape[1] == len(expected_order)
        assert sampling_methods.x_data.shape[1] == len(x_lab)
        assert sampling_methods.data_headers == expected_order
        assert sampling_methods.data_headers_xvars == x_lab
        np.testing.assert_array_equal(
            sampling_methods.data, self.test_data_pandas.values[:, expected_order]
        )
        np.testing.assert_array_equal(
            sampling_methods.x_data, self.test_data_pandas.values[:, x_lab]
        )

    @pytest.mark.unit
    def test_selection_columns_preprocessing_24(self):
        sampling_methods = SamplingMethods()
        x_lab = [2]
        y_lab = [1]
        expected_order = x_lab + y_lab
        sampling_methods.selection_columns_preprocessing(
            self.test_data_pandas.values, x_lab, y_lab
        )
        assert sampling_methods.df_flag == False
        assert sampling_methods.data.shape[1] == len(expected_order)
        assert sampling_methods.x_data.shape[1] == len(x_lab)
        assert sampling_methods.data_headers == expected_order
        assert sampling_methods.data_headers_xvars == x_lab
        np.testing.assert_array_equal(
            sampling_methods.data, self.test_data_pandas.values[:, expected_order]
        )
        np.testing.assert_array_equal(
            sampling_methods.x_data, self.test_data_pandas.values[:, x_lab]
        )

    @pytest.mark.unit
    def test_selection_columns_preprocessing_25(self):
        sampling_methods = SamplingMethods()
        x_lab = [2, 1]
        y_lab = [0, 3]
        expected_order = x_lab + y_lab
        sampling_methods.selection_columns_preprocessing(
            self.test_data_pandas.values, x_lab, y_lab
        )
        assert sampling_methods.df_flag == False
        assert sampling_methods.data.shape[1] == len(expected_order)
        assert sampling_methods.x_data.shape[1] == len(x_lab)
        assert sampling_methods.data_headers == expected_order
        assert sampling_methods.data_headers_xvars == x_lab
        np.testing.assert_array_equal(
            sampling_methods.data, self.test_data_pandas.values[:, expected_order]
        )
        np.testing.assert_array_equal(
            sampling_methods.x_data, self.test_data_pandas.values[:, x_lab]
        )

    @pytest.mark.unit
    def test_selection_columns_preprocessing_26(self):
        sampling_methods = SamplingMethods()
        x_lab = [0]
        y_lab = [4]
        expected_order = x_lab + y_lab

        with pytest.raises(IndexError):
            sampling_methods.selection_columns_preprocessing(
                self.test_data_pandas.values, x_lab, y_lab
            )


class LatinHypercubeSamplingTestCases(unittest.TestCase):
    """
    test__init__selection_01: input numpy array - Test behaviour generate LatinHypercubeSampling object with selection, default number of sample = 5
    test__init__selection_02: input Pandas DataFrame - Test behaviour generate LatinHypercubeSampling object with selection, default number of sample
    test__init__selection_03: input numpy array - Test behaviour generate LatinHypercubeSampling object with selection, with a selected number of sample
    test__init__selection_04: input Pandas DataFrame - Test behaviour generate LatinHypercubeSampling object with selection, with a selected number of sample
    test__init__selection_05: input numpy array - Test behaviour raise exception with a selected number of sample = 0
    test__init__selection_06: input numpy array - Test behaviour raise exception with a selected number of sample = -1
    test__init__selection_07: input numpy array - Test behaviour raise exception with a selected number of sample > input size
    test__init__selection_08: input numpy array - Test behaviour raise exception with a selected number of sample = 1.1 (non-integer)
    test__init__selection_09: input list - Test behaviour raise ValueError with list input

    test__init__creation_01: input list - Test behaviour generate LatinHypercubeSampling object with default sampling_type, default number of sample = 5
    test__init__creation_02: input list - Test behaviour generate LatinHypercubeSampling object with sampling_type = creation, default number of sample = 5
    test__init__creation_03: input list - Test behaviour generate LatinHypercubeSampling object with creation with a selected number of sample
    test__init__creation_04: input list - Test behaviour raise exception with a selected number of sample = 0
    test__init__creation_05: input list - Test behaviour raise exception with a selected number of sample = -1
    test__init__creation_06: input list - Test behaviour raise exception with a selected number of sample = 1.1 (non-integer)
    test__init__creation_07: input numpy - Test behaviour raise ValueError with numpy input
    test__init__creation_08: input numpy - Test behaviour raise ValueError with pandas input

    test__init__creation_selection_01 - Test behaviour raise Exception with sampling_type = non string
    test__init__creation_selection_02 - Test behaviour raise Exception with sampling_type = incorrect string

    test_variable_sample_creation: Test behaviour, sampled values are in the range (min, max), number of samples = 5, 10, 1
    test_lhs_points_generation: Test behaviour, sampled values are in the range (0, 1) , number of samples = 5, 10, 1, 2 d
    test_random_shuffling: Test behaviour, random_shuffling = sampled values after soring, they are in the range (0, 1) , number of samples = 5, 10, 1, 2 d
    test_sample_points_01: Test behaviour with selection, sample points are unique, all in the input array , number of samples = 5, 10, 1, 2 d
    test_sample_points_02: Test behaviour with creation, sample points are unique, all in the input range (min,max) , number of samples = 5, 10, 1, 2 d

    """

    def setUp(self):
        input_array_np = np.array(
            [
                [0, 10, 11],
                [1, 10, 14],
                [2, 10, 19],
                [3, 10, 26],
                [4, 10, 35],
                [5, 10, 46],
                [6, 10, 59],
                [7, 10, 74],
                [8, 10, 91],
                [9, 10, 110],
            ]
        )
        input_array_pd = pd.DataFrame(
            {
                "x1": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                "x2": [10, 10, 10, 10, 10, 10, 10, 10, 10, 10],
                "y": [11, 14, 19, 26, 35, 46, 59, 74, 91, 110],
            }
        )
        input_array_list = [[1, 10, 3], [2, 11, 4.5]]
        self.test_data_numpy = input_array_np
        self.test_data_pandas = input_array_pd
        self.test_data_list = input_array_list

    @pytest.mark.unit
    def test__init__selection_01(self):
        input_array = self.test_data_numpy
        LHSClass = LatinHypercubeSampling(
            input_array, number_of_samples=None, sampling_type="selection"
        )
        np.testing.assert_array_equal(LHSClass.data, input_array)
        np.testing.assert_array_equal(LHSClass.number_of_samples, 5)
        np.testing.assert_array_equal(LHSClass.x_data, input_array[:, :-1])

    @pytest.mark.unit
    def test__init__selection_02(self):
        input_array = self.test_data_pandas
        LHSClass = LatinHypercubeSampling(
            input_array, number_of_samples=None, sampling_type="selection"
        )
        np.testing.assert_array_equal(LHSClass.data, input_array)
        np.testing.assert_array_equal(LHSClass.number_of_samples, 5)
        input_array = input_array.to_numpy()
        np.testing.assert_array_equal(LHSClass.x_data, input_array[:, :-1])

    @pytest.mark.unit
    def test__init__selection_03(self):
        input_array = self.test_data_numpy
        LHSClass = LatinHypercubeSampling(
            input_array, number_of_samples=6, sampling_type="selection"
        )
        np.testing.assert_array_equal(LHSClass.data, input_array)
        np.testing.assert_array_equal(LHSClass.number_of_samples, 6)
        np.testing.assert_array_equal(LHSClass.x_data, input_array[:, :-1])

    @pytest.mark.unit
    def test__init__selection_04(self):
        input_array = self.test_data_pandas
        LHSClass = LatinHypercubeSampling(
            input_array, number_of_samples=6, sampling_type="selection"
        )
        np.testing.assert_array_equal(LHSClass.data, input_array)
        np.testing.assert_array_equal(LHSClass.number_of_samples, 6)
        input_array = input_array.to_numpy()
        np.testing.assert_array_equal(LHSClass.x_data, input_array[:, :-1])

    @pytest.mark.unit
    def test__init__selection_05(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            LHSClass = LatinHypercubeSampling(
                input_array, number_of_samples=0, sampling_type="selection"
            )

    @pytest.mark.unit
    def test__init__selection_06(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            LHSClass = LatinHypercubeSampling(
                input_array, number_of_samples=-1, sampling_type="selection"
            )

    @pytest.mark.unit
    def test__init__selection_07(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            LHSClass = LatinHypercubeSampling(
                input_array, number_of_samples=101, sampling_type="selection"
            )

    @pytest.mark.unit
    def test__init__selection_08(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            LHSClass = LatinHypercubeSampling(
                input_array, number_of_samples=1.1, sampling_type="selection"
            )

    @pytest.mark.unit
    def test__init__selection_09(self):
        input_array = self.test_data_list
        with pytest.raises(ValueError):
            LHSClass = LatinHypercubeSampling(
                input_array, number_of_samples=None, sampling_type="selection"
            )

    @pytest.mark.unit
    def test__init__creation_01(self):
        input_array = self.test_data_list
        LHSClass = LatinHypercubeSampling(
            input_array, number_of_samples=None, sampling_type=None
        )
        np.testing.assert_array_equal(LHSClass.data, input_array)
        np.testing.assert_array_equal(LHSClass.number_of_samples, 5)

    @pytest.mark.unit
    def test__init__creation_02(self):
        input_array = self.test_data_list
        LHSClass = LatinHypercubeSampling(
            input_array, number_of_samples=None, sampling_type="creation"
        )
        np.testing.assert_array_equal(LHSClass.data, input_array)
        np.testing.assert_array_equal(LHSClass.number_of_samples, 5)

    @pytest.mark.unit
    def test__init__creation_03(self):
        input_array = self.test_data_list
        LHSClass = LatinHypercubeSampling(
            input_array, number_of_samples=100, sampling_type="creation"
        )
        np.testing.assert_array_equal(LHSClass.data, input_array)
        np.testing.assert_array_equal(LHSClass.number_of_samples, 100)

    @pytest.mark.unit
    def test__init__creation_04(self):
        input_array = self.test_data_list
        with pytest.raises(Exception):
            LHSClass = LatinHypercubeSampling(
                input_array, number_of_samples=0, sampling_type="creation"
            )

    @pytest.mark.unit
    def test__init__creation_05(self):
        input_array = self.test_data_list
        with pytest.raises(Exception):
            LHSClass = LatinHypercubeSampling(
                input_array, number_of_samples=-1, sampling_type="creation"
            )

    @pytest.mark.unit
    def test__init__creation_06(self):
        input_array = self.test_data_list
        with pytest.raises(Exception):
            LHSClass = LatinHypercubeSampling(
                input_array, number_of_samples=1.1, sampling_type="creation"
            )

    @pytest.mark.unit
    def test__init__creation_07(self):
        input_array = self.test_data_numpy
        with pytest.raises(ValueError):
            LHSClass = LatinHypercubeSampling(
                input_array, number_of_samples=None, sampling_type="creation"
            )

    @pytest.mark.unit
    def test__init__creation_08(self):
        input_array = self.test_data_pandas
        with pytest.raises(ValueError):
            LHSClass = LatinHypercubeSampling(
                input_array, number_of_samples=None, sampling_type="creation"
            )

    @pytest.mark.unit
    def test__init__creation_selection_01(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            LHSClass = LatinHypercubeSampling(
                input_array, number_of_samples=None, sampling_type=1
            )

    @pytest.mark.unit
    def test__init__creation_selection_02(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            LHSClass = LatinHypercubeSampling(
                input_array, number_of_samples=None, sampling_type="jp"
            )

    @pytest.mark.unit
    def test_variable_sample_creation(self):
        input_array = self.test_data_numpy
        for num_samples in [None, 10, 1]:
            LHSClass = LatinHypercubeSampling(
                input_array, number_of_samples=num_samples, sampling_type="selection"
            )
            minimum, maximum = 10, 100
            out_var_samples = LHSClass.variable_sample_creation(minimum, maximum)
            assert (out_var_samples >= minimum).all() and (
                out_var_samples <= maximum
            ).all()
            np.testing.assert_array_equal(
                LHSClass.number_of_samples, out_var_samples.shape[0]
            )

    @pytest.mark.unit
    def test_lhs_points_generation(self):
        input_array = self.test_data_numpy
        for num_samples in [None, 10, 1]:
            LHSClass = LatinHypercubeSampling(
                input_array, number_of_samples=num_samples, sampling_type="selection"
            )
            out_sample_points_vector = LHSClass.lhs_points_generation()
            assert (out_sample_points_vector >= 0).all() and (
                out_sample_points_vector <= 1
            ).all()
            np.testing.assert_array_equal(
                LHSClass.number_of_samples, out_sample_points_vector.shape[0]
            )
            np.testing.assert_array_equal(
                input_array.shape[1] - 1, out_sample_points_vector.shape[1]
            )

    @pytest.mark.unit
    def test_random_shuffling(self):
        input_array = self.test_data_numpy
        for num_samples in [None, 10, 1]:
            LHSClass = LatinHypercubeSampling(
                input_array, number_of_samples=num_samples, sampling_type="selection"
            )
            out_sample_points_vector = LHSClass.lhs_points_generation()
            vector_of_points = LHSClass.random_shuffling(out_sample_points_vector)

            sidx1 = out_sample_points_vector.argsort(axis=0)
            out1 = out_sample_points_vector[sidx1, np.arange(sidx1.shape[1])]
            sidx2 = vector_of_points.argsort(axis=0)
            out2 = vector_of_points[sidx2, np.arange(sidx2.shape[1])]

            assert (out_sample_points_vector >= 0).all() and (
                out_sample_points_vector <= 1
            ).all()
            np.testing.assert_array_equal(out1, out2)
            np.testing.assert_array_equal(
                LHSClass.number_of_samples, out_sample_points_vector.shape[0]
            )
            np.testing.assert_array_equal(
                input_array.shape[1] - 1, out_sample_points_vector.shape[1]
            )

    @pytest.mark.unit
    def test_sample_points_01(self):
        for num_samples in [None, 10, 1]:
            input_array = self.test_data_numpy
            LHSClass = LatinHypercubeSampling(
                input_array, number_of_samples=num_samples, sampling_type="selection"
            )
            unique_sample_points = LHSClass.sample_points()
            expected_testing = np.array(
                [True] * unique_sample_points.shape[0], dtype=bool
            )
            out_testing = [
                unique_sample_points[i, :] in input_array
                for i in range(unique_sample_points.shape[0])
            ]

            np.testing.assert_array_equal(
                np.unique(unique_sample_points, axis=0), unique_sample_points
            )
            np.testing.assert_array_equal(expected_testing, out_testing)

    @pytest.mark.unit
    def test_sample_points_02(self):
        for num_samples in [None, 10, 1]:
            input_array = self.test_data_list
            LHSClass = LatinHypercubeSampling(
                input_array, number_of_samples=num_samples, sampling_type="creation"
            )
            unique_sample_points = LHSClass.sample_points()
            input_array = np.array(input_array)
            for i in range(input_array.shape[1]):
                var_range = input_array[:, i]
                assert (unique_sample_points[:, i] >= var_range[0]).all() and (
                    unique_sample_points[:, i] <= var_range[1]
                ).all()
            np.testing.assert_array_equal(
                np.unique(unique_sample_points, axis=0).shape,
                unique_sample_points.shape,
            )


class UniformSamplingTestCases(unittest.TestCase):
    """
    test__init__selection_01: input numpy array - Test behaviour generate UniformSampling object with selection, default edge, list_of_samples_per_variable = [2,5]
    test__init__selection_02: input Pandas DataFrame - Test behaviour generate UniformSampling object with selection, default edge, list_of_samples_per_variable = [2,5]
    test__init__selection_03: input numpy array - Test behaviour raise TypeError with a list_of_samples_per_variable is numpy, default edge
    test__init__selection_04: input numpy array - Test behaviour raise TypeError with a list_of_samples_per_variable is pandas, default edge
    test__init__selection_05: input numpy array - Test behaviour raise ValueError with a list_of_samples_per_variable < number of variables, default edge
    test__init__selection_06: input numpy array - Test behaviour raise ValueError with a list_of_samples_per_variable > number of variables, default edge
    test__init__selection_07: input numpy array - Test behaviour raise ValueError with a min(list_of_samples_per_variable) < 2, default edge
    test__init__selection_08: input numpy array - Test behaviour raise TypeError with a list_of_samples_per_variable is non integer, default edge
    test__init__selection_09: input numpy array - Test behaviour raise Exception with a list_of_samples_per_variable = [2,50], 2*50 > number of input data, default edge
    test__init__selection_10: input numpy array - Test behaviour generate UniformSampling object with selection, edge = True, list_of_samples_per_variable = [2,5]
    test__init__selection_11: input numpy array - Test behaviour generate UniformSampling object with selection, edge = True, list_of_samples_per_variable = [2,5]
    test__init__selection_12: input numpy array - Test behaviour raise Exception with edge = 1
    test__init__selection_13: input numpy array - Test behaviour raise Exception with edge = 'str'

    test__init__creation_01: input list - Test behaviour generate UniformSampling object with default sampling_type, default edge, list_of_samples_per_variable = [2,7,5]
    test__init__creation_02: input list - Test behaviour generate UniformSampling object with sampling_type = creation, default edge, list_of_samples_per_variable = [2,7,5]
    test__init__creation_03: input list - Test behaviour raise exception with a list_of_samples_per_variable = [1,7,5]
    test__init__creation_04: input list - Test behaviour raise exception with a list_of_samples_per_variable = [-1,7,5]
    test__init__creation_06: input list - Test behaviour raise exception with a list_of_samples_per_variable = [1.1,7,5] (non-integer)
    test__init__creation_07: input numpy - Test behaviour raise ValueError with numpy input
    test__init__creation_08: input Pandas - Test behaviour raise ValueError with Pandas input

    test__init__creation_selection_01 - Test behaviour raise Exception with sampling_type = non string
    test__init__creation_selection_02 - Test behaviour raise Exception with sampling_type = incorrect string

    test_sample_points_01: Test behaviour with selection, sample points are unique, all in the input array , number of samples = 5, 10, 1
    test_sample_points_02: Test behaviour with creation, sample points are unique, all in the input range (min,max) , number of samples = [2,5,9],[3,2,10],[4,2,28]
    """

    def setUp(self):
        input_array_np = np.array(
            [
                [0, 10, 11],
                [1, 10, 14],
                [2, 10, 19],
                [3, 10, 26],
                [4, 10, 35],
                [5, 10, 46],
                [6, 10, 59],
                [7, 10, 74],
                [8, 10, 91],
                [9, 10, 110],
            ]
        )
        input_array_pd = pd.DataFrame(
            {
                "x1": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                "x2": [10, 10, 10, 10, 10, 10, 10, 10, 10, 10],
                "y": [11, 14, 19, 26, 35, 46, 59, 74, 91, 110],
            }
        )
        input_array_list = [[1, 10, 3], [2, 11, 4.5]]
        self.test_data_numpy = input_array_np
        self.test_data_pandas = input_array_pd
        self.test_data_list = input_array_list

    @pytest.mark.unit
    def test__init__selection_01(self):
        input_array = self.test_data_numpy
        UniClass = UniformSampling(input_array, [2, 5], sampling_type="selection")
        np.testing.assert_array_equal(UniClass.data, input_array)
        np.testing.assert_array_equal(UniClass.number_of_samples, 10)
        np.testing.assert_array_equal(UniClass.x_data, input_array[:, :-1])

    @pytest.mark.unit
    def test__init__selection_02(self):
        input_array = self.test_data_pandas
        UniClass = UniformSampling(input_array, [2, 5], sampling_type="selection")
        np.testing.assert_array_equal(UniClass.data, input_array)
        np.testing.assert_array_equal(UniClass.number_of_samples, 10)
        input_array = input_array.to_numpy()
        np.testing.assert_array_equal(UniClass.x_data, input_array[:, :-1])

    @pytest.mark.unit
    def test__init__selection_03(self):
        input_array = self.test_data_numpy
        with pytest.raises(TypeError):
            UniClass = UniformSampling(
                input_array, np.array([2, 5]), sampling_type="selection"
            )

    @pytest.mark.unit
    def test__init__selection_04(self):
        input_array = self.test_data_numpy
        with pytest.raises(TypeError):
            UniClass = UniformSampling(
                input_array, pd.DataFrame([2, 5]), sampling_type="selection"
            )

    @pytest.mark.unit
    def test__init__selection_05(self):
        input_array = self.test_data_numpy
        with pytest.raises(ValueError):
            UniClass = UniformSampling(input_array, [2], sampling_type="selection")

    @pytest.mark.unit
    def test__init__selection_06(self):
        input_array = self.test_data_numpy
        with pytest.raises(ValueError):
            UniClass = UniformSampling(
                input_array, [2, 5, 5], sampling_type="selection"
            )

    @pytest.mark.unit
    def test__init__selection_07(self):
        input_array = self.test_data_numpy
        with pytest.raises(ValueError):
            UniClass = UniformSampling(input_array, [-2, 5], sampling_type="selection")

    @pytest.mark.unit
    def test__init__selection_08(self):
        input_array = self.test_data_numpy
        with pytest.raises(TypeError):
            UniClass = UniformSampling(input_array, [2.1, 5], sampling_type="selection")

    @pytest.mark.unit
    def test__init__selection_09(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            UniClass = UniformSampling(input_array, [2, 50], sampling_type="selection")

    @pytest.mark.unit
    def test__init__selection_10(self):
        input_array = self.test_data_numpy
        UniClass = UniformSampling(
            input_array, [2, 5], sampling_type="selection", edges=True
        )
        np.testing.assert_array_equal(UniClass.data, input_array)
        np.testing.assert_array_equal(UniClass.number_of_samples, 10)
        np.testing.assert_array_equal(UniClass.x_data, input_array[:, :-1])

    @pytest.mark.unit
    def test__init__selection_11(self):
        input_array = self.test_data_numpy
        UniClass = UniformSampling(
            input_array, [2, 5], sampling_type="selection", edges=False
        )
        np.testing.assert_array_equal(UniClass.data, input_array)
        np.testing.assert_array_equal(UniClass.number_of_samples, 10)
        np.testing.assert_array_equal(UniClass.x_data, input_array[:, :-1])

    @pytest.mark.unit
    def test__init__selection_12(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            UniClass = UniformSampling(
                input_array, [2, 5], sampling_type="selection", edges=1
            )

    @pytest.mark.unit
    def test__init__selection_13(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            UniClass = UniformSampling(
                input_array, [2, 5], sampling_type="selection", edges="x"
            )

    @pytest.mark.unit
    def test__init__creation_01(self):
        input_array = self.test_data_list
        UniClass = UniformSampling(input_array, [2, 7, 5], sampling_type=None)
        np.testing.assert_array_equal(UniClass.data, input_array)
        np.testing.assert_array_equal(UniClass.number_of_samples, 2 * 7 * 5)

    @pytest.mark.unit
    def test__init__creation_02(self):
        input_array = self.test_data_list
        UniClass = UniformSampling(input_array, [2, 7, 5], sampling_type="creation")
        np.testing.assert_array_equal(UniClass.data, input_array)
        np.testing.assert_array_equal(UniClass.number_of_samples, 2 * 7 * 5)

    @pytest.mark.unit
    def test__init__creation_03(self):
        input_array = self.test_data_list
        with pytest.raises(Exception):
            UniClass = UniformSampling(input_array, [1, 7, 5], sampling_type="creation")

    @pytest.mark.unit
    def test__init__creation_04(self):
        input_array = self.test_data_list
        with pytest.raises(Exception):
            UniClass = UniformSampling(
                input_array, [-1, 7, 5], sampling_type="creation"
            )

    @pytest.mark.unit
    def test__init__creation_05(self):
        input_array = self.test_data_list
        with pytest.raises(Exception):
            UniClass = UniformSampling(
                input_array, [1.1, 7, 5], sampling_type="creation"
            )

    @pytest.mark.unit
    def test__init__creation_06(self):
        input_array = self.test_data_numpy
        with pytest.raises(ValueError):
            UniClass = UniformSampling(input_array, [2, 5], sampling_type="creation")

    @pytest.mark.unit
    def test__init__creation_07(self):
        input_array = self.test_data_pandas
        with pytest.raises(ValueError):
            UniClass = UniformSampling(input_array, [2, 5], sampling_type="creation")

    @pytest.mark.unit
    def test__init__creation_selection_01(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            UniClass = UniformSampling(input_array, [2, 5], sampling_type=1)

    @pytest.mark.unit
    def test__init__creation_selection_02(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            UniClass = UniformSampling(input_array, [2, 5], sampling_type="jp")

    @pytest.mark.unit
    def test_sample_points_01(self):
        for num_samples in [[2, 5], [3, 2], [4, 2]]:
            input_array = self.test_data_numpy
            UniClass = UniformSampling(
                input_array, num_samples, sampling_type="selection"
            )
            unique_sample_points = UniClass.sample_points()
            expected_testing = np.array(
                [True] * unique_sample_points.shape[0], dtype=bool
            )
            out_testing = [
                unique_sample_points[i, :] in input_array
                for i in range(unique_sample_points.shape[0])
            ]

            np.testing.assert_array_equal(
                np.unique(unique_sample_points, axis=0), unique_sample_points
            )
            np.testing.assert_array_equal(expected_testing, out_testing)

    @pytest.mark.unit
    def test_sample_points_02(self):
        input_array = self.test_data_list
        for num_samples in [[2, 5, 9], [3, 2, 10], [4, 2, 28]]:
            input_array = self.test_data_list
            UniClass = UniformSampling(
                input_array, num_samples, sampling_type="creation"
            )
            unique_sample_points = UniClass.sample_points()
            input_array = np.array(input_array)
            for i in range(input_array.shape[1]):
                var_range = input_array[:, i]
                assert (unique_sample_points[:, i] >= var_range[0]).all() and (
                    unique_sample_points[:, i] <= var_range[1]
                ).all()
            np.testing.assert_array_equal(
                np.unique(unique_sample_points, axis=0).shape,
                unique_sample_points.shape,
            )


class HammersleySamplingTestCases(unittest.TestCase):
    """
    __init__ = __init__ in LatinHypercubeSampling except Dimensionality problem:
        test__init__selection: Test behaviour with  dimensionality > 10

    test_sample_points_01: Test behaviour with selection, sample points are unique, all in the input array , number of samples = 5, 10, 1
    test_sample_points_02: Test behaviour with creation, sample points are unique, all in the input range (min,max) , number of samples = 5, 10, 1

    """

    def setUp(self):
        input_array_np = np.array(
            [
                [0, 10, 11],
                [1, 10, 14],
                [2, 10, 19],
                [3, 10, 26],
                [4, 10, 35],
                [5, 10, 46],
                [6, 10, 59],
                [7, 10, 74],
                [8, 10, 91],
                [9, 10, 110],
            ]
        )
        input_array_np_large = np.array(
            [
                [0, 10, 11, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                [1, 10, 14, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                [2, 10, 19, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                [3, 10, 26, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                [4, 10, 35, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                [5, 10, 46, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                [6, 10, 59, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                [7, 10, 74, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                [8, 10, 91, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                [9, 10, 110, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
            ]
        )
        input_array_pd = pd.DataFrame(
            {
                "x1": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                "x2": [10, 10, 10, 10, 10, 10, 10, 10, 10, 10],
                "y": [11, 14, 19, 26, 35, 46, 59, 74, 91, 110],
            }
        )
        input_array_list = [[1, 10, 3], [2, 11, 4.5]]
        self.test_data_numpy = input_array_np
        self.test_data_numpy_large = input_array_np_large
        self.test_data_pandas = input_array_pd
        self.test_data_list = input_array_list

    @pytest.mark.unit
    def test__init__selection(self):
        input_array = self.test_data_numpy_large
        with pytest.raises(Exception):
            HammersleyClass = HammersleySampling(
                input_array, number_of_samples=None, sampling_type="selection"
            )

    @pytest.mark.unit
    def test_sample_points_01(self):
        for num_samples in [None, 10, 1]:
            input_array = self.test_data_numpy
            HammersleyClass = HammersleySampling(
                input_array, number_of_samples=num_samples, sampling_type="selection"
            )
            unique_sample_points = HammersleyClass.sample_points()
            expected_testing = np.array(
                [True] * unique_sample_points.shape[0], dtype=bool
            )
            out_testing = [
                unique_sample_points[i, :] in input_array
                for i in range(unique_sample_points.shape[0])
            ]

            np.testing.assert_array_equal(
                np.unique(unique_sample_points, axis=0), unique_sample_points
            )
            np.testing.assert_array_equal(expected_testing, out_testing)

    @pytest.mark.unit
    def test_sample_points_02(self):
        for num_samples in [None, 10, 1]:
            input_array = self.test_data_list
            HammersleyClass = HammersleySampling(
                input_array, number_of_samples=num_samples, sampling_type="creation"
            )
            unique_sample_points = HammersleyClass.sample_points()
            input_array = np.array(input_array)
            for i in range(input_array.shape[1]):
                var_range = input_array[:, i]
                assert (unique_sample_points[:, i] >= var_range[0]).all() and (
                    unique_sample_points[:, i] <= var_range[1]
                ).all()
            np.testing.assert_array_equal(
                np.unique(unique_sample_points, axis=0).shape,
                unique_sample_points.shape,
            )


class CVTSamplingTestCases(unittest.TestCase):
    """
    test__init__selection_01: input numpy array - Test behaviour generate object with selection, default number of sample = 5, default tolerance
    test__init__selection_02: input Pandas DataFrame - Test behaviour generate object with selection, default number of sample, default tolerance
    test__init__selection_03: input numpy array - Test behaviour generate object with selection, with a selected number of sample, default tolerance
    test__init__selection_04: input Pandas DataFrame - Test behaviour generate object with selection, with a selected number of sample, default tolerance
    test__init__selection_05: input numpy array - Test behaviour raise exception with a selected number of sample = 0, default tolerance
    test__init__selection_06: input numpy array - Test behaviour raise exception with a selected number of sample = -1, default tolerance
    test__init__selection_07: input numpy array - Test behaviour raise exception with a selected number of sample > input size, default tolerance
    test__init__selection_08: input numpy array - Test behaviour raise exception with a selected number of sample = 1.1 (non-integer), default tolerance
    test__init__selection_09: input list - Test behaviour raise ValueError with list input
    test__init__selection_10: input numpy array - Test behaviour raise exception with tolerance > 0.1
    test__init__selection_11: input numpy array - Test behaviour raise Warning with tolerance < 0.9
    test__init__selection_12: input numpy array - Test behaviour tolerance   0.09
    test__init__selection_13: input numpy array - Test behaviour tolerance  -0.09

    test__init__creation_01: input list - Test behaviour generate object with default sampling_type, default number of sample = 5, default tolerance
    test__init__creation_02: input list - Test behaviour generate object with sampling_type = creation, default number of sample = 5, default tolerance
    test__init__creation_03: input list - Test behaviour generate object with creation with a selected number of sample, default tolerance
    test__init__creation_04: input list - Test behaviour raise exception with a selected number of sample = 0
    test__init__creation_05: input list - Test behaviour raise exception with a selected number of sample = -1
    test__init__creation_06: input list - Test behaviour raise exception with a selected number of sample = 1.1 (non-integer)
    test__init__creation_07: input numpy - Test behaviour raise ValueError with numpy input
    test__init__creation_08: input numpy - Test behaviour raise ValueError with pandas input
    test__init__creation_09: input list - Test behaviour raise exception with tolerance > 0.1
    test__init__creation_10: input list - Test behaviour raise Warning with tolerance < 0.9
    test__init__creation_11: input list - Test behaviour tolerance   0.09
    test__init__creation_12: input list - Test behaviour tolerance  -0.09

    test__init__creation_selection_01 - Test behaviour raise Exception with sampling_type = non string
    test__init__creation_selection_02 - Test behaviour raise Exception with sampling_type = incorrect string


    test_random_sample_selection_01 - Test random_sample_selection with size = (5,2)
    test_random_sample_selection_02 - Test random_sample_selection with size = (0,2)
    test_random_sample_selection_03 - Test random_sample_selection with size = (2,0)
    test_random_sample_selection_04 - Test behaviour raise ValueError with  size = (5,-1)
    test_random_sample_selection_05 - Test behaviour raise TypeError with size =(5,1.1)

    Tests: : Unit tests for eucl_distance, a function that evaluates the distance between two points (u, v).
        Four demonstration tests are done:
    test_eucl_distance_01 - The first test checks that the correct result is obtained when both inputs are single value arrays.
    test_eucl_distance_02 - The second test checks that the correct result is returned when both inputs are 2D vectors.
    test_eucl_distance_03 - The third test checks that the correct result is returned when both inputs are arrays of the same size.
    test_eucl_distance_04 - The fourth test checks that the function is able to calculate the distance from a single point (n x 1 row vector) to a set of design points (supplied in an n x m array)

    Unit tests for create_centres, a function that generates new mass centroids for the design space based on McQueen's method.
        Four demonstration tests are done:
    test_create_centres_01 - The first test checks that the correct result for the new centres is returned when the counter is at its lowest value (1).
    test_create_centres_02 - The second test checks that the correct result for the new centres is returned when the counter is at an arbitrary value (10).
    test_create_centres_03 - The third test checks that the correct procedure is followed when one of the centres in initial_centres has no close design point to it.
    test_create_centres_04 - The fourth test checks that the the approach works as expected for problems with more than two dimensions.

    test_sample_points_01: Test behaviour with selection, sample points are unique, all in the input array , number of samples = 5, 10, 1
    test_sample_points_02: Test behaviour with creation, sample points are unique, all in the input range (min,max) , number of samples = 5, 10, 1

    """

    def setUp(self):
        input_array_np = np.array(
            [
                [0, 10, 11],
                [1, 10, 14],
                [2, 10, 19],
                [3, 10, 26],
                [4, 10, 35],
                [5, 10, 46],
                [6, 10, 59],
                [7, 10, 74],
                [8, 10, 91],
                [9, 10, 110],
            ]
        )
        input_array_pd = pd.DataFrame(
            {
                "x1": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                "x2": [10, 10, 10, 10, 10, 10, 10, 10, 10, 10],
                "y": [11, 14, 19, 26, 35, 46, 59, 74, 91, 110],
            }
        )
        input_array_list = [[1, 10, 3], [2, 11, 4.5]]
        self.test_data_numpy = input_array_np
        self.test_data_pandas = input_array_pd
        self.test_data_list = input_array_list

    @pytest.mark.unit
    def test__init__selection_01(self):
        input_array = self.test_data_numpy
        CVTClass = CVTSampling(
            input_array,
            number_of_samples=None,
            tolerance=None,
            sampling_type="selection",
        )
        np.testing.assert_array_equal(CVTClass.data, input_array)
        np.testing.assert_array_equal(CVTClass.number_of_centres, 5)
        np.testing.assert_array_equal(CVTClass.x_data, input_array[:, :-1])
        np.testing.assert_array_equal(CVTClass.eps, 1e-7)

    @pytest.mark.unit
    def test__init__selection_02(self):
        input_array = self.test_data_pandas
        CVTClass = CVTSampling(
            input_array,
            number_of_samples=None,
            tolerance=None,
            sampling_type="selection",
        )
        np.testing.assert_array_equal(CVTClass.data, input_array)
        np.testing.assert_array_equal(CVTClass.number_of_centres, 5)
        input_array = input_array.to_numpy()
        np.testing.assert_array_equal(CVTClass.x_data, input_array[:, :-1])
        np.testing.assert_array_equal(CVTClass.eps, 1e-7)

    @pytest.mark.unit
    def test__init__selection_03(self):
        input_array = self.test_data_numpy
        CVTClass = CVTSampling(
            input_array, number_of_samples=6, tolerance=None, sampling_type="selection"
        )
        np.testing.assert_array_equal(CVTClass.data, input_array)
        np.testing.assert_array_equal(CVTClass.number_of_centres, 6)
        np.testing.assert_array_equal(CVTClass.x_data, input_array[:, :-1])
        np.testing.assert_array_equal(CVTClass.eps, 1e-7)

    @pytest.mark.unit
    def test__init__selection_04(self):
        input_array = self.test_data_pandas
        CVTClass = CVTSampling(
            input_array, number_of_samples=6, tolerance=None, sampling_type="selection"
        )
        np.testing.assert_array_equal(CVTClass.data, input_array)
        np.testing.assert_array_equal(CVTClass.number_of_centres, 6)
        input_array = input_array.to_numpy()
        np.testing.assert_array_equal(CVTClass.x_data, input_array[:, :-1])
        np.testing.assert_array_equal(CVTClass.eps, 1e-7)

    @pytest.mark.unit
    def test__init__selection_05(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            CVTClass = CVTSampling(
                input_array,
                number_of_samples=0,
                tolerance=None,
                sampling_type="selection",
            )

    @pytest.mark.unit
    def test__init__selection_06(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            CVTClass = CVTSampling(
                input_array,
                number_of_samples=-1,
                tolerance=None,
                sampling_type="selection",
            )

    @pytest.mark.unit
    def test__init__selection_07(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            CVTClass = CVTSampling(
                input_array,
                number_of_samples=101,
                tolerance=None,
                sampling_type="selection",
            )

    @pytest.mark.unit
    def test__init__selection_08(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            CVTClass = CVTSampling(
                input_array,
                number_of_samples=1.1,
                tolerance=None,
                sampling_type="selection",
            )

    @pytest.mark.unit
    def test__init__selection_09(self):
        input_array = self.test_data_list
        with pytest.raises(ValueError):
            CVTClass = CVTSampling(
                input_array,
                number_of_samples=None,
                tolerance=None,
                sampling_type="selection",
            )

    @pytest.mark.unit
    def test__init__selection_10(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            CVTClass = CVTSampling(
                input_array,
                number_of_samples=None,
                tolerance=0.11,
                sampling_type="selection",
            )

    @pytest.mark.unit
    def test__init__selection_11(self):
        input_array = self.test_data_numpy
        warning_msg = (
            "Tolerance too tight. CVT algorithm may take long time to converge."
        )
        with self.assertLogs(level="WARNING") as cm:
            CVTClass = CVTSampling(
                input_array,
                number_of_samples=None,
                tolerance=1e-10,
                sampling_type="selection",
            )
        self.assertIn(warning_msg, "\n".join(cm.output))
        for record in cm.records:
            self.assertEqual(record.levelno, idaeslog.WARNING)

    @pytest.mark.unit
    def test__init__selection_12(self):
        input_array = self.test_data_numpy
        CVTClass = CVTSampling(
            input_array,
            number_of_samples=None,
            tolerance=0.09,
            sampling_type="selection",
        )
        np.testing.assert_array_equal(CVTClass.eps, 0.09)

    @pytest.mark.unit
    def test__init__selection_13(self):
        input_array = self.test_data_numpy
        CVTClass = CVTSampling(
            input_array,
            number_of_samples=None,
            tolerance=-0.09,
            sampling_type="selection",
        )
        np.testing.assert_array_equal(CVTClass.eps, -0.09)

    @pytest.mark.unit
    def test__init__creation_01(self):
        input_array = self.test_data_list
        CVTClass = CVTSampling(
            input_array, number_of_samples=None, tolerance=None, sampling_type=None
        )
        np.testing.assert_array_equal(CVTClass.data, input_array)
        np.testing.assert_array_equal(CVTClass.number_of_centres, 5)

    @pytest.mark.unit
    def test__init__creation_02(self):
        input_array = self.test_data_list
        CVTClass = CVTSampling(
            input_array,
            number_of_samples=None,
            tolerance=None,
            sampling_type="creation",
        )
        np.testing.assert_array_equal(CVTClass.data, input_array)
        np.testing.assert_array_equal(CVTClass.number_of_centres, 5)

    @pytest.mark.unit
    def test__init__creation_03(self):
        input_array = self.test_data_list
        CVTClass = CVTSampling(
            input_array, number_of_samples=100, tolerance=None, sampling_type="creation"
        )
        np.testing.assert_array_equal(CVTClass.data, input_array)
        np.testing.assert_array_equal(CVTClass.number_of_centres, 100)

    @pytest.mark.unit
    def test__init__creation_04(self):
        input_array = self.test_data_list
        with pytest.raises(Exception):
            CVTClass = CVTSampling(
                input_array,
                number_of_samples=0,
                tolerance=None,
                sampling_type="creation",
            )

    @pytest.mark.unit
    def test__init__creation_05(self):
        input_array = self.test_data_list
        with pytest.raises(Exception):
            CVTClass = CVTSampling(
                input_array,
                number_of_samples=-1,
                tolerance=None,
                sampling_type="creation",
            )

    @pytest.mark.unit
    def test__init__creation_06(self):
        input_array = self.test_data_list
        with pytest.raises(Exception):
            CVTClass = CVTSampling(
                input_array,
                number_of_samples=1.1,
                tolerance=None,
                sampling_type="creation",
            )

    @pytest.mark.unit
    def test__init__creation_07(self):
        input_array = self.test_data_numpy
        with pytest.raises(ValueError):
            CVTClass = CVTSampling(
                input_array,
                number_of_samples=None,
                tolerance=None,
                sampling_type="creation",
            )

    @pytest.mark.unit
    def test__init__creation_08(self):
        input_array = self.test_data_pandas
        with pytest.raises(ValueError):
            CVTClass = CVTSampling(
                input_array,
                number_of_samples=None,
                tolerance=None,
                sampling_type="creation",
            )

    @pytest.mark.unit
    def test__init__creation_09(self):
        input_array = self.test_data_list
        with pytest.raises(Exception):
            CVTClass = CVTSampling(
                input_array,
                number_of_samples=None,
                tolerance=0.11,
                sampling_type="creation",
            )

    @pytest.mark.unit
    def test__init__creation_10(self):
        warning_msg = (
            "Tolerance too tight. CVT algorithm may take long time to converge."
        )
        input_array = self.test_data_list
        with self.assertLogs(level="WARNING") as cm:
            CVTClass = CVTSampling(
                input_array,
                number_of_samples=None,
                tolerance=1e-10,
                sampling_type="creation",
            )
        self.assertIn(warning_msg, "\n".join(cm.output))
        for record in cm.records:
            assert record.levelno == idaeslog.WARNING

    @pytest.mark.unit
    def test__init__creation_11(self):
        input_array = self.test_data_list
        CVTClass = CVTSampling(
            input_array,
            number_of_samples=None,
            tolerance=0.09,
            sampling_type="creation",
        )
        np.testing.assert_array_equal(CVTClass.eps, 0.09)

    @pytest.mark.unit
    def test__init__creation_11(self):
        input_array = self.test_data_list
        CVTClass = CVTSampling(
            input_array,
            number_of_samples=None,
            tolerance=-0.09,
            sampling_type="creation",
        )
        np.testing.assert_array_equal(CVTClass.eps, -0.09)

    @pytest.mark.unit
    def test__init__creation_selection_01(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            CVTClass = CVTSampling(
                input_array, number_of_samples=None, tolerance=None, sampling_type=1
            )

    @pytest.mark.unit
    def test__init__creation_selection_02(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            CVTClass = LatinHypercubeSampling(
                input_array, number_of_samples=None, tolerance=None, sampling_type="jp"
            )

    @pytest.mark.unit
    def test_random_sample_selection_01(self):
        size = (5, 2)
        out_random_points = CVTSampling.random_sample_selection(size[0], size[1])
        assert (out_random_points >= 0).all() and (out_random_points <= 1).all()
        assert out_random_points.shape == size

    @pytest.mark.unit
    def test_random_sample_selection_02(self):
        size = (0, 2)
        out_random_points = CVTSampling.random_sample_selection(size[0], size[1])
        assert (out_random_points >= 0).all() and (out_random_points <= 1).all()
        assert out_random_points.shape == size

    @pytest.mark.unit
    def test_random_sample_selection_03(self):
        size = (2, 0)
        out_random_points = CVTSampling.random_sample_selection(size[0], size[1])
        assert (out_random_points >= 0).all() and (out_random_points <= 1).all()
        assert out_random_points.shape == size

    @pytest.mark.unit
    def test_random_sample_selection_04(self):
        size = (5, -1)
        with pytest.raises(ValueError):
            out_random_points = CVTSampling.random_sample_selection(size[0], size[1])

    @pytest.mark.unit
    def test_random_sample_selection_05(self):
        size = (5, 1.1)
        with pytest.raises(TypeError):
            out_random_points = CVTSampling.random_sample_selection(size[0], size[1])

    @pytest.mark.unit
    def test_eucl_distance_01(self):
        u = np.array([[3]])
        v = np.array([[5]])
        expected_output = 2
        output = CVTSampling.eucl_distance(u, v)
        assert expected_output == output

    @pytest.mark.unit
    def test_eucl_distance_02(self):
        u = np.array([[1, 2]])
        v = np.array([[3, 4]])
        expected_output = 8**0.5
        output = CVTSampling.eucl_distance(u, v)
        assert expected_output == output

    @pytest.mark.unit
    def test_eucl_distance_03(self):
        u = np.array([[1, 2], [3, 4]])
        v = np.array([[5, 6], [7, 8]])
        expected_output = np.array([32**0.5, 32**0.5])
        output = CVTSampling.eucl_distance(u, v)
        np.testing.assert_array_equal(expected_output, output)

    @pytest.mark.unit
    def test_eucl_distance_04(self):
        u = np.array([[1, 2]])
        v = np.array([[5, 6], [7, 8]])
        expected_output = np.array([32**0.5, 72**0.5])
        output = CVTSampling.eucl_distance(u, v)
        np.testing.assert_array_equal(expected_output, output)

    @pytest.mark.unit
    def test_create_centres_01(self):
        initial_centres = np.array([[0, 0], [1, 1]])
        current_random_points = np.array([[0.6, 0.6], [0.3, 0.3]])
        current_centres = np.array([1, 0])
        counter = 1
        expected_output = np.array([[0.15, 0.15], [0.8, 0.8]])
        output = CVTSampling.create_centres(
            initial_centres, current_random_points, current_centres, counter
        )
        np.testing.assert_array_equal(expected_output, output)

    @pytest.mark.unit
    def test_create_centres_02(self):
        initial_centres = np.array([[0, 0], [1, 1]])
        current_random_points = np.array([[0.6, 0.6], [0.3, 0.3]])
        current_centres = np.array([1, 0])
        counter = 10
        expected_output = np.array([[0.3 / 11, 0.3 / 11], [10.6 / 11, 10.6 / 11]])
        output = CVTSampling.create_centres(
            initial_centres, current_random_points, current_centres, counter
        )
        np.testing.assert_array_equal(expected_output, output)

    @pytest.mark.unit
    def test_create_centres_03(self):
        initial_centres = np.array([[0, 0], [1, 1]])
        current_random_points = np.array([[0.6, 0.6], [0.8, 0.8]])
        current_centres = np.array([1, 1])
        counter = 5
        expected_output = np.array([[0.5 / 6, 0.5 / 6], [5.7 / 6, 5.7 / 6]])
        output = CVTSampling.create_centres(
            initial_centres, current_random_points, current_centres, counter
        )
        np.testing.assert_array_equal(expected_output, output)

    @pytest.mark.unit
    def test_create_centres_04(self):
        initial_centres = np.array([[0, 0, 0], [1, 1, 1]])
        current_random_points = np.array(
            [
                [0.1, 0.1, 0.1],
                [0.3, 0.3, 0.3],
                [0.5, 0.5, 0.5],
                [0.7, 0.7, 0.7],
                [0.9, 0.9, 0.9],
            ]
        )
        current_centres = np.array([0, 0, 0, 1, 1])
        counter = 4
        expected_output = np.array(
            [[0.3 / 5, 0.3 / 5, 0.3 / 5], [4.8 / 5, 4.8 / 5, 4.8 / 5]]
        )
        output = CVTSampling.create_centres(
            initial_centres, current_random_points, current_centres, counter
        )
        np.testing.assert_array_equal(expected_output, output)

    @pytest.mark.component
    def test_sample_points_01(self):
        for num_samples in [None, 10, 1]:
            input_array = self.test_data_numpy
            CVTClass = CVTSampling(
                input_array,
                number_of_samples=num_samples,
                tolerance=None,
                sampling_type="selection",
            )
            unique_sample_points = CVTClass.sample_points()
            expected_testing = np.array(
                [True] * unique_sample_points.shape[0], dtype=bool
            )
            out_testing = [
                unique_sample_points[i, :] in input_array
                for i in range(unique_sample_points.shape[0])
            ]

            np.testing.assert_array_equal(
                np.unique(unique_sample_points, axis=0), unique_sample_points
            )
            np.testing.assert_array_equal(expected_testing, out_testing)

    @pytest.mark.component
    def test_sample_points_02(self):
        for num_samples in [None, 10, 1]:
            input_array = self.test_data_list
            CVTClass = CVTSampling(
                input_array, number_of_samples=num_samples, sampling_type="creation"
            )
            unique_sample_points = CVTClass.sample_points()
            input_array = np.array(input_array)
            for i in range(input_array.shape[1]):
                var_range = input_array[:, i]
                assert (unique_sample_points[:, i] >= var_range[0]).all() and (
                    unique_sample_points[:, i] <= var_range[1]
                ).all()
            np.testing.assert_array_equal(
                np.unique(unique_sample_points, axis=0).shape,
                unique_sample_points.shape,
            )


if __name__ == "__main__":
    unittest.main()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import pytest

import pyomo.common.unittest as unittest
import pyomo.environ as pyo
from pyomo.dae.flatten import flatten_dae_components
from pyomo.network.arc import Arc

from pyomo.contrib.incidence_analysis import (
    IncidenceGraphInterface,
)
from pyomo.util.check_units import assert_units_consistent

import idaes.core as idaes
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    large_residuals_set,
)
from idaes.models_extra.gas_distribution.properties.natural_gas import (
    NaturalGasParameterBlock,
)
from idaes.models_extra.gas_distribution.unit_models.pipeline import GasPipeline
from idaes.models_extra.gas_distribution.unit_models.compressor import (
    IsothermalCompressor as Compressor,
)

from idaes.apps.nmpc.dynamic_data import (
    load_inputs_into_model,
    interval_data_from_time_series,
)
from idaes.core.solvers import get_solver


@pytest.mark.component
class TestSolveDynamicPipelineCompressor(unittest.TestCase):
    def make_steady_model(
        self,
        nfe=2,
        scheme="FORWARD",
    ):
        m = pyo.ConcreteModel()
        default = {"dynamic": False}
        m.fs = idaes.FlowsheetBlock(**default)
        m.fs.properties = NaturalGasParameterBlock()
        pipeline_config = {
            "property_package": m.fs.properties,
            "finite_elements": nfe,
            "transformation_scheme": scheme,
            "has_holdup": True,
        }
        m.fs.pipeline = GasPipeline(**pipeline_config)
        pipeline = m.fs.pipeline
        compressor_config = {"property_package": m.fs.properties}
        m.fs.compressor = Compressor(**compressor_config)
        compressor = m.fs.compressor
        cv = m.fs.pipeline.control_volume
        # Fix geometry variables
        m.fs.pipeline.diameter.fix(0.92 * pyo.units.m)
        cv.length.fix(300.0 * pyo.units.km)

        m._compressor_to_pipeline = Arc(
            ports=(compressor.outlet_port, pipeline.inlet_port),
        )
        expand_arcs = pyo.TransformationFactory("network.expand_arcs")
        expand_arcs.apply_to(m)
        return m

    def fix_model_inlets(
        self,
        model,
        inlet_flow_mass=3.0e5 * pyo.units.kg / pyo.units.hr,
        inlet_pressure=57.0 * pyo.units.bar,
        inlet_temperature=293.15 * pyo.units.K,
    ):
        cv = model.fs.pipeline.control_volume
        j = next(iter(model.fs.properties.component_list))
        x0 = cv.length_domain.first()
        state = model.fs.compressor.inlet_state
        state[:].mole_frac_comp[j].fix()
        state[:].temperature.fix(inlet_temperature)
        state[:].pressure.fix(inlet_pressure)
        cv.flow_mass[:, x0].fix(inlet_flow_mass)

    def get_scalar_data_from_model(
        self,
        model,
        time,
        scalar_vars=None,
        dae_vars=None,
    ):
        if scalar_vars is None or dae_vars is None:
            scalar_vars, dae_vars = flatten_dae_components(model, time, pyo.Var)
        return {str(pyo.ComponentUID(var)): var.value for var in scalar_vars}

    def get_data_from_model_at_time(
        self,
        model,
        time,
        scalar_vars=None,
        dae_vars=None,
        t0=0,
    ):
        if scalar_vars is None or dae_vars is None:
            scalar_vars, dae_vars = flatten_dae_components(model, time, pyo.Var)
        return {str(pyo.ComponentUID(var.referent)): var[t0].value for var in dae_vars}

    def test_sim(self):
        """
        Inlet pressure and outlet flow rate will be fixed.
        """
        nxfe = 4
        ipopt = get_solver("ipopt")

        m_steady = self.make_steady_model(nfe=nxfe)
        self.fix_model_inlets(m_steady, inlet_pressure=50.0 * pyo.units.bar)
        m_steady.fs.compressor.boost_pressure[:].fix(7.0 * pyo.units.bar)
        ipopt.solve(m_steady, tee=True)
        time_steady = m_steady.fs.time
        scalar_data = self.get_scalar_data_from_model(m_steady, time_steady)
        initial_data = self.get_data_from_model_at_time(m_steady, time_steady)

        m = pyo.ConcreteModel()
        default = {
            "dynamic": True,
            "time_set": [0.0, 20.0],
            "time_units": pyo.units.hr,
        }
        m.fs = idaes.FlowsheetBlock(**default)
        m.fs.properties = NaturalGasParameterBlock()
        pipeline_config = {
            "property_package": m.fs.properties,
            "finite_elements": nxfe,
        }
        m.fs.pipeline = GasPipeline(**pipeline_config)
        pipeline = m.fs.pipeline
        compressor_config = {"property_package": m.fs.properties}
        m.fs.compressor = Compressor(**compressor_config)
        compressor = m.fs.compressor
        m._compressor_to_pipeline = Arc(
            ports=(compressor.outlet_port, pipeline.inlet_port),
        )
        expand_arcs = pyo.TransformationFactory("network.expand_arcs")
        expand_arcs.apply_to(m)

        cv = m.fs.pipeline.control_volume
        assert_units_consistent(m)

        disc = pyo.TransformationFactory("dae.finite_difference")
        ntfe = 20
        disc.apply_to(m, nfe=ntfe, wrt=m.fs.time, scheme="BACKWARD")

        time = m.fs.time
        t0 = m.fs.time.first()
        x0 = cv.length_domain.first()
        xf = cv.length_domain.last()
        j = next(iter(m.fs.properties.component_list))

        # Fix geometry variables
        m.fs.pipeline.diameter.fix(0.92 * pyo.units.m)
        cv.length.fix(300.0 * pyo.units.km)

        # Fix boost pressure
        compressor.boost_pressure[:].fix()

        # Inlets to the compressor are fixed, except for flow, where
        # the outlet is fixed.
        state = compressor.inlet_state
        state[:].pressure.fix()
        state[:].mole_frac_comp[j].fix()
        state[:].temperature.fix()
        cv.flow_mass[:, xf].fix()

        # Fix initial conditions. Here, pressure and volume for all
        # non-specified points.
        for x in cv.length_domain:
            if x != x0:
                cv.pressure[t0, x].fix()
            if x != xf:
                cv.flow_mass[t0, x].fix()

        # I want to deactivate differential equations at (t0, xf)
        # Material balance already doesn't exist here.
        cv.momentum_balance[t0, xf].deactivate()

        self.assertEqual(degrees_of_freedom(m), 0)

        # Load initial steady state into model at all time points.
        for name, val in initial_data.items():
            var = m.find_component(name)
            for t in time:
                var[t].set_value(val)
        # Load scalar data from initial steady state
        # (initialize area, basically)
        for name, val in scalar_data.items():
            var = m.find_component(name)
            var.set_value(val)

        cv.material_accumulation[...].set_value(0.0)
        cv.flow_mass_dt[...].set_value(0.0)

        for con in large_residuals_set(m):
            resid = pyo.value(con.body - con.upper)
            print(resid, con.name)
        ipopt.solve(m, tee=True)

        # Load input sequence into model
        sample_points = [4.0, 20.0]
        input_name = "fs.pipeline.control_volume.flow_mass[*,1.0]"
        nominal_density = 0.72
        val = 12.0 * 1e6 / 24 * nominal_density  # 12 (1e6 SCM)/day
        input_series_data = (
            sample_points,
            {input_name: [val, val]},
        )
        input_interval_data = interval_data_from_time_series(input_series_data)
        load_inputs_into_model(m, time, input_interval_data)
        # Solve with loaded inputs
        res = ipopt.solve(m, tee=True)
        self.assertIs(
            res.solver.termination_condition,
            pyo.TerminationCondition.optimal,
        )

        # These predicted values come from a simulation of a single pipeline
        # model from the Pyomo DAE example. flow_mass has been converted
        # to kg/hr from (1e4 SCM/hr) by a factor of 0.72*1e4, where
        # 0.72 kg/m**3 is the gas density at standard conditions.
        pred_values = (
            list(time),
            {
                "fs.pipeline.control_volume.flow_mass[*,%s]"
                % x0: [
                    3.000e5,
                    2.999e5,
                    2.999e5,
                    2.999e5,
                    3.000e5,
                    3.174e5,
                    3.301e5,
                    3.389e5,
                    3.449e5,
                    3.492e5,
                    3.523e5,
                    3.544e5,
                    3.560e5,
                    3.571e5,
                    3.579e5,
                    3.585e5,
                    3.589e5,
                    3.592e5,
                    3.594e5,
                    3.595e5,
                    3.597e5,
                ],
                "fs.pipeline.control_volume.pressure[*,%s]"
                % xf: [
                    50.90,
                    50.90,
                    50.90,
                    50.90,
                    50.90,
                    49.83,
                    49.31,
                    48.95,
                    48.69,
                    48.51,
                    48.38,
                    48.29,
                    48.22,
                    48.17,
                    48.14,
                    48.11,
                    48.10,
                    48.08,
                    48.07,
                    48.07,
                    48.06,
                ],
                "fs.compressor.power[*]": [
                    1.590e3,
                    1.590e3,
                    1.590e3,
                    1.590e3,
                    1.590e3,
                    1.682e3,
                    1.750e3,
                    1.796e3,
                    1.828e3,
                    1.851e3,
                    1.867e3,
                    1.878e3,
                    1.887e3,
                    1.892e3,
                    1.897e3,
                    1.900e3,
                    1.902e3,
                    1.904e3,
                    1.905e3,
                    1.906e3,
                    1.906e3,
                ],
            },
        )
        output_names = [
            "fs.pipeline.control_volume.flow_mass[*,%s]" % x0,
            "fs.pipeline.control_volume.pressure[*,%s]" % xf,
            "fs.compressor.power[*]",
        ]
        actual_values = (
            list(time),
            {
                name: [var.value for var in m.find_component(name).values()]
                for name in output_names
            },
        )
        # Note: We fail with a reltol of 0.01, due to flow rate discrepancies
        # in positions 6, 7, 8, and 9. A reltol of 0.02 seems reasonable to me.
        self.assertStructuredAlmostEqual(pred_values, actual_values, reltol=0.02)


@pytest.mark.unit
class TestConstructPipelineCompressorFlowsheet(unittest.TestCase):
    """
    Test for construction of a flowsheet consisting of one compressor
    feeding into one pipeline. This is equivalent to an "active link"
    in V. Zavala's 2014 paper.
    """

    def test_steady(self):
        """ """
        m = pyo.ConcreteModel()
        default = {
            "dynamic": False,
        }
        m.fs = idaes.FlowsheetBlock(**default)
        m.fs.properties = NaturalGasParameterBlock()
        pipeline_config = {
            "property_package": m.fs.properties,
            "finite_elements": 2,
        }
        m.fs.pipeline = GasPipeline(**pipeline_config)
        pipeline = m.fs.pipeline
        compressor_default = {"property_package": m.fs.properties}
        m.fs.compressor = Compressor(**compressor_default)
        compressor = m.fs.compressor

        m._compressor_to_pipeline = Arc(
            ports=(compressor.outlet_port, pipeline.inlet_port),
        )
        expand_arcs = pyo.TransformationFactory("network.expand_arcs")
        expand_arcs.apply_to(m)

        cv = m.fs.pipeline.control_volume
        p = "Vap"
        j = next(iter(m.fs.properties.component_list))
        t = m.fs.time.first()
        x0 = cv.length_domain.first()

        # Fix pipeline degrees of freedom:
        # Design variables:
        m.fs.pipeline.diameter.fix(0.92 * pyo.units.m)
        m.fs.pipeline.control_volume.length.fix(300.0 * pyo.units.m)

        # Fix compressor degrees of freedom:
        compressor.boost_pressure[:].fix()
        # Inlet variables:
        state = m.fs.compressor.inlet_state
        state[:].mole_frac_comp[j].fix(1.0)
        state[:].temperature.fix(300.0 * pyo.units.K)
        state[:].pressure.fix(57.0 * pyo.units.bar)

        # With just the pipeline unit model, we've been fixing to
        # 3e5 kg/hr. Need to convert this into kmol/hr
        ng_comp = m.fs.properties.get_component(j)
        # Here I'm assuming that we have a single component
        state[:].flow_mol.fix(3.0e5 * pyo.units.kg / pyo.units.hr / ng_comp.mw)

        # This test asserts:
        # (a) consistent units
        # (b) zero degrees of freedom
        # (c) a perfect matching between constraints and variables
        assert_units_consistent(m)
        self.assertEqual(degrees_of_freedom(m), 0)
        igraph = IncidenceGraphInterface(m)
        matching = igraph.maximum_matching()
        N, M = igraph.incidence_matrix.shape
        self.assertEqual(N, M)
        self.assertEqual(len(matching), N)

    def test_dynamic(self):
        """ """
        m = pyo.ConcreteModel()
        default = {
            "dynamic": True,
            "time_set": [0.0, 20.0],
            "time_units": pyo.units.hr,
        }
        m.fs = idaes.FlowsheetBlock(**default)
        m.fs.properties = NaturalGasParameterBlock()
        pipeline_config = {
            "property_package": m.fs.properties,
            "finite_elements": 4,
        }
        m.fs.pipeline = GasPipeline(**pipeline_config)
        pipeline = m.fs.pipeline

        compressor_config = {"property_package": m.fs.properties}
        m.fs.compressor = Compressor(**compressor_config)
        compressor = m.fs.compressor

        m._pipeline_to_compressor = Arc(
            ports=(compressor.outlet_port, pipeline.inlet_port),
        )
        expand_arcs = pyo.TransformationFactory("network.expand_arcs")
        expand_arcs.apply_to(m)

        p = "Vap"
        j = next(iter(m.fs.properties.component_list))

        # Assert units consistent before discretization transformation
        assert_units_consistent(m)
        disc = pyo.TransformationFactory("dae.finite_difference")
        ntfe = 5
        disc.apply_to(m, nfe=ntfe, wrt=m.fs.time, scheme="BACKWARD")

        cv = m.fs.pipeline.control_volume
        t0 = m.fs.time.first()
        x0 = cv.length_domain.first()
        xf = cv.length_domain.last()

        # Fix geometry variables
        m.fs.pipeline.diameter.fix(0.92 * pyo.units.m)
        cv.length.fix(300.0 * pyo.units.km)

        # Fix boost pressure
        compressor.boost_pressure[:].fix()

        # Fix inlet and outlet degrees of freedom
        state = compressor.inlet_state
        state[:].pressure.fix()
        state[:].mole_frac_comp[j].fix()
        state[:].temperature.fix()
        cv.flow_mass[:, xf].fix()

        # Fix initial conditions. Here, pressure and volume for all
        # non-inlet points.
        for x in cv.length_domain:
            if x != x0:
                cv.pressure[t0, x].fix()
            if x != xf:
                cv.flow_mass[t0, x].fix()

        # I want to deactivate differential equations at (t0, xf)
        # Material balance already doesn't exist here.
        cv.momentum_balance[t0, xf].deactivate()

        igraph = IncidenceGraphInterface(m)
        N, M = igraph.incidence_matrix.shape
        matching = igraph.maximum_matching()
        self.assertEqual(degrees_of_freedom(m), 0)
        self.assertEqual(N, M)  # Sanity check
        self.assertEqual(len(matching), N)


if __name__ == "__main__":
    unittest.main()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
IDAES Parameter Sweep API and sequential workflow runner.
"""

import sys
import json

from pandas import DataFrame

from pyomo.core import Param, Var
from pyomo.environ import check_optimal_termination
from pyomo.common.config import ConfigDict, ConfigValue, document_kwargs_from_configdict

import idaes.logger as idaeslog
from idaes.core.surrogate.pysmo.sampling import SamplingMethods, UniformSampling
from idaes.core.util.exceptions import ConfigurationError

__author__ = "Andrew Lee"

# Set up logger
_log = idaeslog.getLogger(__name__)

# TODO: Timeouts
# TODO: Re-initialize option/callback


class ParameterSweepSpecification(object):
    """Defines a set of input variables/parameters and values to be used in
    a parameter sweep study.

    Users can define inputs to be sampled, the number of samples and the method to
    use to generate the samples (chosen from Pysmo's sampling methods) and then
    call the generate_samples method to generate the desired set of samples.
    """

    # TODO: Consider supporting sampling from data sets in the future
    def __init__(self):
        self._inputs = {}
        self._sampling_method = None
        self._samples = None
        self._sample_size = None

    def add_sampled_input(
        self,
        pyomo_path: str,
        lower: float,
        upper: float,
        name: str = None,
    ):
        """
        Add an input that should be sampled when forming the set of
        specific points that need to be run for the convergence evaluation

        The input will be sampled assuming a normal distribution
        (with given mean and standard deviation)
        truncated to the values given by lower and upper bounds

        Args:
            pyomo_path A string representation of the path to the variable or parameter to
                be sampled. This string will be executed to retrieve the Pyomo
            component.
            lower : lower bound on the input variable or parameter.
            upper : upper bound on the input variable or parameter
            name : (Optional) display name for the input. If not provided,
                pyomo_pth will be used.

        Returns:
            None
        """
        # ToDo: put some error checking here ... Maybe we should have the model
        # ToDo: already? Can use to check if the pyomo_path is valid? check if
        # ToDo: bounds will be violated?
        spec = {}
        spec["pyomo_path"] = pyomo_path
        spec["lower"] = lower
        spec["upper"] = upper

        if name is None:
            # If no name given, use Pyomo path
            name = pyomo_path

        self._inputs[name] = spec

    def set_sampling_method(self, sampling_method):
        """Defines the method to use to generate samples. This must be a Pysmo
        SamplingMethod

        Args:
            sampling_method: SamplingMethod class to use.

        Returns:
            None
        """
        try:
            if issubclass(sampling_method, SamplingMethods):
                self._sampling_method = sampling_method
            else:
                raise TypeError
        except TypeError:
            raise TypeError(
                f"Sampling method must be an instance of a Pysmo SamplingMethod "
                f"(received {sampling_method})"
            )

    def set_sample_size(self, sample_size):
        """Sets the number of samples to be generated.

        The format depends on the sampling method chosen. For UniformSampling, this must
        be a list of ints of length equal to the number of inputs, otherwise it must be
        an int.

        Args:
            sample_size: int or list-of-ints

        Returns:
            None
        """
        self._sample_size = sample_size

    def _generate_pysmo_data_input(self):
        lb_list = []
        ub_list = []

        for var_def in self.inputs.values():
            lb_list.append(var_def["lower"])
            ub_list.append(var_def["upper"])

        return [lb_list, ub_list]

    def generate_samples(self):
        """
        Generate set of samples of specified size from inputs.

        Args:
            sample_size - [int or list of ints] size of sample to be generated.
                For UniformSampling this must be a list of ints indicating number of
                graduations for each input. For all other sampling methods this must be
                an int indicating the number of samples to generate.

        Returns:
            Pandas DataFrame of samples
        """
        if self.sampling_method is None:
            raise ValueError(
                "Please choose a sampling method to use for sample generation. "
                "Sampling method must be chosen from those available in Pysmo."
            )

        if len(self._inputs) == 0:
            raise ValueError("Please identify at least one input variable to sample.")

        if self._sample_size is None:
            raise ValueError("Please set a sample size.")
        elif self.sampling_method is UniformSampling:
            if not isinstance(self._sample_size, list):
                raise TypeError(
                    "For UniformSampling, sample_size must be list of integers."
                )
            sample_size = self._sample_size
        else:
            try:
                sample_size = int(self._sample_size)
            except ValueError:
                raise ValueError("sample_size must be an integer.")
            if sample_size <= 0:
                raise ValueError("sample_size must be an integer greater than 1.")

        bounds_list = self._generate_pysmo_data_input()

        # pylint: disable=not-callable
        space_init = self.sampling_method(
            bounds_list,
            sample_size,
            sampling_type="creation",
        )

        self._samples = DataFrame(
            space_init.sample_points(), columns=[_ for _ in self.inputs]
        )

        return self.samples

    @property
    def inputs(self):
        """
        Returns an dict containing the declared inputs.
        """
        return self._inputs

    @property
    def sampling_method(self):
        """
        Returns the declared sampling method.
        """
        return self._sampling_method

    @property
    def sample_size(self):
        """
        Returns the declared sample size.
        """
        return self._sample_size

    @property
    def samples(self):
        """
        Returns the generated set of samples (pandas DataFrame).
        """
        return self._samples

    def to_dict(self):
        """Write specification to a dict

        Returns:
            Input specification as a dict
        """
        outdict = {}

        outdict["inputs"] = self._inputs

        # Pysmo sampling methods are not serializable, so get name instead
        outdict["sampling_method"] = self._sampling_method.__name__
        outdict["sample_size"] = self._sample_size
        outdict["samples"] = self._samples.to_dict(orient="tight")

        return outdict

    def from_dict(self, input_dict: dict):
        """
        Load specification from dict

        Args:
            input_dict: dict to load into specification

        Returns:
            None
        """
        self._inputs = {}
        for k, v in input_dict["inputs"].items():
            self._inputs[k] = v

        # Hack to get Pysom sampling method from string name
        mod = __import__(
            "idaes.core.surrogate.pysmo.sampling",
            fromlist=[input_dict["sampling_method"]],
        )
        self._sampling_method = getattr(mod, input_dict["sampling_method"])

        self._sample_size = input_dict["sample_size"]
        self._samples = DataFrame().from_dict(
            input_dict["samples"],
            orient="tight",
        )

    def to_json_file(self, filename: str):
        """
        Serialize specification to file in json format.

        Args:
            filename: name of file to write to as string

        Returns:
            None
        """
        with open(filename, "w") as fd:
            json.dump(self.to_dict(), fd, indent=3)

    def from_json_file(self, filename: str):
        """
        Load specification from json file.

        Args:
            filename: name of file to load as string

        Returns:
            None
        """
        with open(filename, "r") as f:
            self.from_dict(json.load(f))
        f.close()


def is_psweepspec(val):
    """
    Config validator for ParameterSweepSpecifications

    Args:
        val: value to validate

    Returns:
        ParameterSweepSpecification

    Raises:
        ValueError if val is not a ParameterSweepSpecification
    """
    if isinstance(val, ParameterSweepSpecification):
        return val
    _log.error(
        f"Input configuration {val} must be an instance of ParameterSweepSpecification."
    )
    raise ValueError(
        "Input configuration must be an instance of ParameterSweepSpecification."
    )


def _is_solver(val):
    """
    Config validator for Solver object.

    Use duck-typing and just look for a solve method

    Args:
        val: value to validate

    Returns:
        val if it has a solve method

    Raises:
        ValueError if val does not have a solve method
    """
    try:
        if callable(val.solve):
            return val
    except AttributeError:
        raise ValueError("solver must be an instance of a Pyomo Solver object.")


CONFIG = ConfigDict()
CONFIG.declare(
    "rebuild_model",
    ConfigValue(
        default=True,
        domain=bool,
        doc="Whether to rebuild model for each sample, or reuse the same instance for "
        "all runs (default=True, rebuild for all runs).",
    ),
)
CONFIG.declare(
    "build_model",
    ConfigValue(doc="Callback method to construct initialized model for execution."),
)
CONFIG.declare(
    "build_model_arguments",
    ConfigValue(
        domain=dict,
        doc="Arguments to pass to build_model callback.",
    ),
)
CONFIG.declare(
    "run_model",
    ConfigValue(
        doc="Callback method to use when running model. If None, model is run with solver.solve()"
    ),
)
CONFIG.declare(
    "run_model_arguments",
    ConfigValue(
        domain=dict,
        doc="Arguments to pass to run_model callback.",
    ),
)
CONFIG.declare(
    "build_outputs",
    ConfigValue(
        doc="Callback method to use to collect results from model after execution."
    ),
)
CONFIG.declare(
    "build_outputs_arguments",
    ConfigValue(
        domain=dict,
        doc="Arguments to pass to build_outputs callback.",
    ),
)
CONFIG.declare(
    "handle_solver_error",
    ConfigValue(
        doc="Callback method to use in case of exception when solving a sample."
    ),
)
CONFIG.declare(
    "handle_solver_error_arguments",
    ConfigValue(
        domain=dict,
        doc="Arguments to pass to handle_solver_error callback.",
    ),
)
CONFIG.declare(
    "halt_on_error",
    ConfigValue(
        default=False,
        domain=bool,
        doc="Whether to halt execution of parameter sweep on encountering a solver error (default=False).",
    ),
)
CONFIG.declare(
    "input_specification",
    ConfigValue(
        domain=is_psweepspec,
        doc="ParameterSweepSpecification object defining inputs to be sampled",
    ),
)
CONFIG.declare(
    "solver",
    ConfigValue(
        default=None,
        domain=_is_solver,
        doc="Pyomo solver object to use when solving model",
    ),
)


@document_kwargs_from_configdict(CONFIG)
class ParameterSweepBase:
    """
    IDAES base class for defining parameter sweep runners.

    Thus base class defines a basic API for setting up parameter sweep studies
    and allows the end-user to provide instructions for how to set up and run
    the model to be studied using a set of callbacks.
    """

    def __init__(self, **kwargs):
        self.config = CONFIG(kwargs)
        self._results = {}
        self._model = None  # used to store model instance if rebuild_model is False

    @property
    def results(self):
        """
        Returns dict containing the results from the parameter sweep.
        """
        return self._results

    def execute_parameter_sweep(self):
        """
        Placeholder method for parameter sweep runners.

        Developers should overload this with a method to execute the parameter sweep
        using their preferred workflow manager.

        Raises:
            NotImplementedError
        """
        raise NotImplementedError(
            "Derived classes should overload this method with a workflow manager."
        )

    def execute_single_sample(self, sample_id: int):
        """
        Executes a single run of the model using input values from sample_id.

        Args:
            sample_id: int indicating row in specification.samples to load into model
                for run.

        Returns:
            results: results generated by build_outputs callback
            success: bool indicating whether execution was successful
            error: str if error occurs during solve else None
        """
        model = self.get_initialized_model()

        # Load sample values
        self.set_input_values(model, sample_id)

        # Try/except to catch any critical failures that occur
        error = None
        try:
            success, run_stats = self.run_model(model, self.config.solver)

            if not success:
                _log.warning(f"Sample {sample_id} did not report success.")
        except Exception as e:  # pylint: disable=broad-except
            if self.config.halt_on_error:
                raise

            success = False
            error = str(e)  # Cast to string for storage

        # Compile Results
        if error is None:
            results = self.build_outputs(model, run_stats)
        else:
            # Catch any Exception for recourse
            results = self.handle_error(model)

        _log.info(f"Sample {sample_id} finished.")

        return results, success, error

    def get_initialized_model(self):
        """
        Get instance of model to be run by calling build_model callback.

        Returns:
            Pyomo model to be run
        """
        if self.config.build_model is None:
            raise ConfigurationError(
                "Please specify a method to construct the model of interest."
            )

        if not self.config.rebuild_model:
            # If reusing model, see if instance has been constructed yet
            if self._model is not None:
                # If yes, return and done
                return self._model

        # Otherwise, build instance of model
        args = self.config.build_model_arguments
        if args is None:
            args = {}
        model = self.config.build_model(**args)
        # TODO: Verify model is actually a model?

        if not self.config.rebuild_model:
            # If reusing model, store instance for reuse
            self._model = model

        return model

    def get_input_specification(self):
        """
        Get input specification from config block

        Returns:
            ParameterSweepConfiguration object

        Raises:
            ConfigurationError if no input specification has been defined
        """
        if self.config.input_specification is None:
            raise ConfigurationError(
                "Please specify an input specification to use for sampling."
            )

        return self.config.input_specification

    def get_input_samples(self):
        """
        Get dataframe of samples from input specification.

        Returns:
            pandas dataframe of samples

        Raises:
            ConfigurationError if no input specification has been defined
        """
        spec = self.get_input_specification()

        samples = spec.samples

        if samples is None:
            samples = spec.generate_samples()

        return samples

    def get_sample_values(self, sample_id: int):
        """
        Get inputs for a specific sample run indicated by sample_id.

        Args:
            sample_id: int representing a row in the specification.samples dataframe

        Returns:
            Pandas Series of input values for chosen sample
        """
        samples = self.get_input_samples()

        return samples.iloc[sample_id]

    def set_input_values(self, model, sample_id: int):
        """
        Set values of input variables/parameters in instance of model using values
        from sample_id.

        Args:
            model: instance of model to be executed
            sample_id: int representing a row in the specification.samples dataframe

        Returns:
            None

        Raises:
            ValueError if an input cannot be found or if it is not a fixed Var or
            mutable Param
        """
        samples = self.get_input_samples()
        inputs = self.config.input_specification.inputs

        for k, i in inputs.items():
            v = samples[k][sample_id]

            # k stores the "name" of the input
            # need to get the pyomo path of the input from the inputs structure
            pyomo_path = i["pyomo_path"]

            comp = model.find_component(pyomo_path)
            try:
                ctype = comp.ctype
            except AttributeError:
                ctype = None

            # TODO: Validate bounds

            if ctype is Param:
                if comp.is_constant():
                    raise ValueError(
                        f"Convergence testing found an input of type Param that "
                        f"was not mutable ({comp.name}). Please make sure all "
                        f"sampled inputs are either mutable params or fixed vars."
                    )
                comp.set_value(v)
            elif ctype is Var:
                try:
                    if not comp.is_fixed():
                        raise ValueError(
                            f"Convergence testing found an input of type Var that "
                            f"was not fixed ({comp.name}). Please make sure all "
                            f"sampled inputs are either mutable params or fixed vars."
                        )
                    comp.set_value(float(v))
                except AttributeError:
                    # Component might be indexed, try iterating and setting value
                    for i, c in comp.items():
                        if not c.is_fixed():
                            raise ValueError(
                                f"Convergence testing found an input of type IndexedVar that "
                                f"was not fixed ({comp.name}, index {i}). Please make sure all "
                                f"sampled inputs are either mutable params or fixed vars."
                            )
                        c.set_value(float(v))
            else:
                raise ValueError(
                    f"Failed to find a valid input component (must be "
                    f"a fixed Var or a mutable Param). Instead, "
                    f"pyomo_path: {pyomo_path} returned: {comp}."
                )

    def run_model(self, model, solver):
        """
        Executes run of model by calling the run_model callback

        Args:
            model: instance of model to be run
            solver: Pyomo solver object to use to solve model

        Returns:
            success: bool indicating whether execution was successful or not
            run_stats: output collected by run_model callback (default is Pyomo SolverResults object)
        """
        if self.config.run_model is None:
            res = solver.solve(model)

            success = check_optimal_termination(res)
            return success, res

        args = self.config.run_model_arguments
        if args is None:
            args = {}

        return self.config.run_model(model, solver, **args)

    def build_outputs(self, model, run_stats):
        """
        Collects desired results from instance of model by calling build_outputs callback.

        Args:
            model: instance of model to collect results from
            run_stats: output from run_model callback to be collected in results
                (default is Pyomo results object)

        Returns:
            Output of build_outputs callback. If no build_outputs callback is provided,
            returns run_stats instead.
        """
        if self.config.build_outputs is None:
            return run_stats
        else:
            args = self.config.build_outputs_arguments
            if args is None:
                args = {}

            return self.config.build_outputs(model, run_stats, **args)

    def handle_error(self, model):
        """
        Call handle_solver_error callback. This method is used in case the solver encounters
        a critical error whilst solving a sample.

        Args:
            model: instance of model for performing recourse

        Returns:
            Output of handle_solver_error callback
        """
        if self.config.handle_solver_error is None:
            # No recourse specified, so results=None
            return None
        else:
            args = self.config.handle_solver_error_arguments
            if args is None:
                args = {}
            return self.config.handle_solver_error(model, **args)

    def to_dict(self):
        """
        Serialize specification and current results to dict form

        Returns:
            dict
        """
        # TODO : Need to serialize build_model method somehow?
        outdict = {}
        outdict["specification"] = self.get_input_specification().to_dict()
        outdict["results"] = self.results

        return outdict

    def from_dict(self, input_dict: dict):
        """
        Load specification and results from dict.

        Args:
            input_dict: dict to load from

        Returns:
            None
        """
        if self.config.input_specification is not None:
            # Log a warning about overwriting
            _log.debug("Overwriting existing input specification")

        self.config.input_specification = ParameterSweepSpecification()
        self.config.input_specification.from_dict(input_dict["specification"])

        self._results = {}
        # Need to iterate to convert string indices to int
        # Converting to json turns the indices to strings
        for k, v in input_dict["results"].items():
            self._results[int(k)] = v

    def to_json_file(self, filename: str):
        """
        Write specification and results to json file.

        Args:
            filename: name of file to write to as string

        Returns:
            None
        """
        with open(filename, "w") as fd:
            json.dump(self.to_dict(), fd, indent=3)

    def from_json_file(self, filename: str):
        """
        Load specification and results from json file.

        Args:
            filename: name of file to load from as string

        Returns:
            None
        """
        with open(filename, "r") as f:
            self.from_dict(json.load(f))
        f.close()

    @staticmethod
    def progress_bar(fraction, msg: str, length: int = 20):
        """
        Prints a progress bar to stdout

        Args:
            fraction: fraction of total samples executed
            msg: string to append to progress bar
            length: length of progress bar (default=20)

        Returns:
            None
        """
        n_complete = int(length * fraction)
        n_remaining = length - n_complete
        characters = f"{'*' * n_complete}{'-' * n_remaining}"
        sys.stdout.write(f"{fraction * 100:.1f}% {characters} {msg}\n")
        sys.stdout.flush()


class SequentialSweepRunner(ParameterSweepBase):
    """
    Sequential runner for parameter sweeps.

    This class executes a parameter sweep by running all samples sequentially
    in a for loop.
    """

    def execute_parameter_sweep(self):
        """
        Execute sequential parameter sweep.

        Returns:
            dict of results indexed by sample ID.
        """
        self._results = {}
        samples = self.get_input_samples()

        count = 1
        for s in samples.index:
            sresults, success, error = self.execute_single_sample(s)
            self._results[s] = {"success": success, "results": sresults, "error": error}

            self.progress_bar(float(count) / float(len(samples)), "Complete")
            count += 1

        return self.results
